import numpy as np
from scipy.ndimage import distance_transform_edt
from skimage.feature import peak_local_max
from skimage.filters import gaussian


def origins_from_distance(array):
    """
    Compute a list of local peaks (origins) in the input array.
    The function applies a Gaussian filter to the input array, creates a binary
    mask using a threshold based on the standard deviation of the filtered data, and
    performs a distance transform on the mask. It finds the local peaks of the
    distance-transformed image with a minimum distance between the peaks,
    and returns the coordinates and a map of the peaks.

    Args:
        array (np.array): Input 2D numpy array.

    Returns:
        peak_list (np.array): List of peak coordinates, shape (num_peaks, 2).
        peak_map (np.array): Binary map of peaks, same shape as input array.

    Example:
    >>> array = np.random.rand(10, 10)
    >>> peak_list, peak_map = origins_from_distance(array)
    """

    # Apply Gaussian filter to the input array
    array = gaussian(array, sigma=1)
    std = np.std(array.flatten())

    # Threshold the filtered array and perform distance transform
    thr_mask = np.where(array > 3 * std, 1, 0)
    distance = distance_transform_edt(thr_mask)

    # Find local peaks in the distance-transformed array
    peak_list = peak_local_max(distance, min_distance=5, threshold_abs=std, footprint=np.ones((3, 3)))

    # Create a map with True elements at peak locations
    peak_map = np.zeros_like(array, dtype=bool)  ### WHAT IS BASE AND WHAT DOES IT DO??
    peak_map[tuple(peak_list.T)] = True

    return peak_list, peak_map



import numpy as np
from numpy.typing import NDArray
from ome_zarr.io import parse_url
from ome_zarr.reader import Reader


def read_ome_zarr(path: str, magnification: str = "0", array: NDArray | None = None) -> NDArray | None:
    """Reads an OME-Zarr file from a given path.

    Args:
        path: Path to the OME-Zarr file
        magnification: Magnification level to be read
        array: Optional numpy array to store the image data. If None, returns a new array

    Returns:
        The image data as a numpy array if array is None, otherwise None after updating
        the provided array

    Example:
        >>> image = read_ome_zarr("path/to/file.zarr")
        >>> # Or with existing array:
        >>> existing_array = np.zeros((100, 100))
        >>> read_ome_zarr("path/to/file.zarr", array=existing_array)
    """
    # read the image data
    loc = parse_url(path, mode="r")
    zarr_reader = Reader(loc).zarr

    # read entire data into memory
    if array is None:
        image = np.array(zarr_reader.load(magnification).compute())
        return image
    else:
        array = np.array(zarr_reader.load(magnification).compute())
        return None



from pathlib import PosixPath
from typing import TypeVar

import yaml

T = TypeVar("T")


def read_config(config_path: str | PosixPath) -> dict:
    with open(config_path) as stream:
        try:
            config = yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            print(exc)
    return config


def flatten(nested_list: list[list[T]]) -> list[T | tuple[T]]:
    """Flatten a list of lists into a single list.

    Args:
        nested_list: A list containing one or more lists as its elements

    Returns:
        A single list containing all elements from the input lists

    Example:
        >>> nested_list = [[1, 2, 3], [4, 5], [6, 7, 8, 9]]
        >>> flatten(nested_list)
        [1, 2, 3, 4, 5, 6, 7, 8, 9]
    """
    return [item for sublist in nested_list for item in sublist]



from pathlib import Path
from typing import Literal

from scportrait.data._dataloader import _download
from scportrait.tools.ml.pretrained_models import get_data_dir


def _get_remote_dataset(
    dataset: str,
    url: str,
    name: str | None = None,
    archive_format: Literal["zip", "tar", "tar.gz", "tgz"] | None = "zip",
    outfile_name: None | str = None,
) -> Path:
    """Download and extract a dataset from a remote location.

    Args:
        dataset: Name of the folder to save the dataset
        url: URL to the dataset
        name: Name of a specific file in the dataset folder to return, if None the whole dataset folder is returned

    Returns:
        Path to the downloaded and extracted dataset or a specific file in the dataset
    """
    data_dir = get_data_dir()
    save_path = data_dir / dataset
    if not save_path.exists():
        _download(url=url, output_path=str(save_path), output_file_name=outfile_name, archive_format=archive_format)
    if name is None:
        return save_path
    else:
        return save_path / name


def custom_cellpose_model() -> Path:
    """Download the example custom cellpose model.

    Returns:
        Path to the downloaded and extracted custom cellpose model
    """
    DATASET = "custom_cellpose_model"
    URL = "https://zenodo.org/records/14931602/files/custom_cellpose_model.cpkt?download=1"
    NAME = "custom_cellpose_model.cpkt"
    return _get_remote_dataset(DATASET, URL, NAME, archive_format=None, outfile_name=NAME)


def dataset_1() -> Path:
    """Download and extract the example dataset 1 images.

    Returns:
        Path to the downloaded and extracted images.
    """
    DATASET = "example_1_images"
    URL = "https://zenodo.org/records/13385933/files/example_1_images.zip?download=1"
    return _get_remote_dataset(DATASET, URL)


def dataset_1_omezarr() -> Path:
    """Download and extract the example dataset 1 images in ome.zarr format.

    Returns:
        Path to the downloaded ome.zarr file.
    """
    DATASET = "example_1_images_omezarr"
    URL = "https://zenodo.org/records/14841309/files/input_image.ome.zarr.zip?download=1"
    NAME = "input_image.ome.zarr"
    return _get_remote_dataset(DATASET, URL, NAME)


def dataset_2() -> Path:
    """Download and extract the example dataset 2 images.

    Returns:
        Path to the downloaded and extracted images.
    """
    DATASET = "example_2_images"
    URL = "https://zenodo.org/records/13742316/files/example_2_images.zip?download=1"
    return _get_remote_dataset(DATASET, URL)


def dataset_3() -> Path:
    """Download and extract the example dataset 3 images.

    Returns:
        Path to the downloaded and extracted images.
    """
    DATASET = "example_3_images"
    URL = "https://zenodo.org/records/13742319/files/example_3_images.zip?download=1"
    return _get_remote_dataset(DATASET, URL)


def dataset_4() -> Path:
    """Download and extract the example dataset 4 images.

    Returns:
        Path to the downloaded and extracted images.
    """
    DATASET = "example_4_images"
    URL = "https://zenodo.org/records/13742331/files/example_4_images.zip?download=1"
    return _get_remote_dataset(DATASET, URL)


def dataset_5() -> Path:
    """Download and extract the example dataset 5 images.

    Returns:
        Path to the downloaded and extracted images.
    """
    DATASET = "example_5_images"
    URL = "https://zenodo.org/records/13742344/files/example_5_images.zip?download=1"
    return _get_remote_dataset(DATASET, URL)


def dataset_6() -> Path:
    """Download and extract the example dataset 6 images.

    Returns:
        Path to the downloaded and extracted images.
    """
    DATASET = "example_6_images"
    URL = "https://zenodo.org/records/13742373/files/example_6_images.zip?download=1"
    return _get_remote_dataset(DATASET, URL)


def dataset_7() -> Path:
    """Download and extract the example dataset 7 images.

    Returns:
        Path to the downloaded and extracted images.
    """
    DATASET = "example_7_images"
    URL = "https://zenodo.org/records/14951047/files/example_7_images.zip?download=1"
    return _get_remote_dataset(DATASET, URL)


def dataset_stitching_example() -> Path:
    """Download and extract the example dataset for stitching images.

    Returns:
        Path to the downloaded and extracted images.
    """
    DATASET = "stitching_example"
    URL = "https://zenodo.org/records/13742379/files/example_stitching.zip?download=1"
    return _get_remote_dataset(DATASET, URL)


def dataset_parsing_example() -> Path:
    """Download and extract the example dataset for parsing Harmony exported imaging experiments.

    Returns:
        Path to the downloaded and extracted images.
    """
    DATASET = "parsing_example_basic"
    URL = "https://zenodo.org/records/14193689/files/harmony_export_V7_basic.zip?download=1"
    NAME = "basic_export"
    return _get_remote_dataset(DATASET, URL, NAME)


def dataset_parsing_example_flatfield_corrected() -> Path:
    """Download and extract the example dataset for parsing Harmony exported imaging experiments.

    Returns:
        Path to the downloaded and extracted images.
    """
    DATASET = "parsing_example_flatfield"
    URL = "https://zenodo.org/records/14973295/files/harmony_export_V7_flatfield_corrected.zip?download=1"
    NAME = "flat_field_corrected"
    return _get_remote_dataset(DATASET, URL, NAME)



import os

import numpy as np
import skimage.exposure
from ashlar import filepattern
from ashlar.filepattern import FilePatternReader
from ashlar.reg import BioformatsMetadata, BioformatsReader
from skimage.filters import gaussian
from skimage.util import invert

from scportrait.processing.images._image_processing import rescale_image


class FilePatternReaderRescale(FilePatternReader):
    """Class for reading images based on a file pattern. If desired the images can be rescaled to a certain range while reading."""

    def __init__(
        self,
        path: str,
        pattern: str,
        overlap: float,
        pixel_size: int = 1,
        do_rescale: bool = False,
        no_rescale_channel: str = "Alexa488",
        rescale_range: tuple[int, int] = (1, 99),
        dtype=None,
    ):
        try:
            super().__init__(path, pattern, overlap, pixel_size=pixel_size)
        except (OSError, FileNotFoundError):
            print(
                f"Error: Could not read images with the given pattern {pattern}. Please check the path {path} and pattern."
            )
            found_files = os.listdir(path)
            print(
                f"At the provided location the the files follow the naming convention:{found_files[0:max(5, len(found_files))]} "
            )

        self.do_rescale = do_rescale
        self.no_rescale_channel = no_rescale_channel
        self.rescale_range = rescale_range
        self.dtype_image = dtype

    @staticmethod
    def rescale(
        img: np.ndarray, rescale_range: tuple[int, int] = (1, 99), cutoff_threshold: int | None = None
    ) -> np.ndarray:
        return rescale_image(img, rescale_range, cutoff_threshold=cutoff_threshold)

    @staticmethod
    # placeholer method kept for compatibility with old code
    # should be reimplemented in the future to allow for more flexible illumination correction
    def correct_illumination(
        img: np.ndarray,
        sigma: float = 30,
        double_correct: bool = False,
        rescale_range: tuple[int, int] = (1, 99),
        cutoff_threshold: int | None = None,
    ) -> np.ndarray:
        img = rescale_image(img, rescale_range, cutoff_threshold=cutoff_threshold, return_float=True)

        # calculate correction mask
        correction = gaussian(img, sigma)
        correction = invert(correction)
        correction = skimage.exposure.rescale_intensity(correction, out_range=(0, 1))

        correction_lows = np.where(img > 0.5, 0, img) * correction
        img_corrected = skimage.exposure.rescale_intensity(img + correction_lows, out_range=(0, 1))

        if double_correct:
            correction_mask_highs = invert(correction)
            correction_mask_highs_02 = skimage.exposure.rescale_intensity(
                np.where(img_corrected < 0.5, 0, img_corrected) * correction_mask_highs
            )
            img_corrected_double = skimage.exposure.rescale_intensity(img_corrected - 0.25 * correction_mask_highs_02)

            return (img_corrected_double * 65535).astype("uint16")
        else:
            return (img_corrected * 65535).astype("uint16")

    def read(self, series, c) -> np.ndarray:
        img = super().read(series, c)

        # check to ensure we are always using the same dtype
        if self.dtype_image is not None:
            if not isinstance(img.dtype, self.dtype_image):
                Warning(f"Found image with dtype {img.dtype}. Automatically converting to {self.dtype_image}")
                img = img.astype(self.dtype_image)
        else:
            # first time reading an image set dtype if none is supplied
            print("Setting dtype to", img.dtype)
            self.dtype_image = img.dtype

        if not self.do_rescale:
            return img
        else:
            if c not in self.no_rescale_channel:
                # get rescale_range for channel c
                if isinstance(self.rescale_range, dict):
                    rescale_range = self.rescale_range[c]
                else:
                    rescale_range = self.rescale_range

                # actually read the image with rescaling applied
                return self.rescale(img, rescale_range=rescale_range)
            else:
                return img


class BioformatsMetadataRescale(BioformatsMetadata):
    """Reimplementation of BioformatsMetadata class to provide same parameters as contained in FilePatternReaderRescale"""

    def __init__(self, path):
        super().__init__(path)

    @property
    def channel_map(self):
        n_channels = self._metadata.getChannelCount(0)
        channel_names = []

        for id in range(n_channels):
            channel_names.append(self._metadata.getChannelName(0, id))
        channel_map = dict(zip(list(range(n_channels)), channel_names, strict=False))

        return channel_map


class BioformatsReaderRescale(BioformatsReader):
    """Class for reading images from Bioformats files (e.g. nd2).

    If desired the images can be rescaled to a certain range while reading.
    """

    def __init__(self, path, plate=None, well=None, do_rescale=False, no_rescale_channel=None, rescale_range=(1, 99)):
        self.path = path
        self.metadata = BioformatsMetadataRescale(self.path)
        self.metadata.set_active_plate_well(plate, well)
        self.do_rescale = do_rescale
        self.no_rescale_channel = no_rescale_channel
        self.rescale_range = rescale_range

    @staticmethod
    def rescale(img, rescale_range=(1, 99)):
        img = skimage.util.img_as_float32(img)
        cutoff1, cutoff2 = rescale_range

        if img.max() > (40000 / 65535):
            _img = img.copy()
            _img[_img > (10000 / 65535)] = 0
            p1 = np.percentile(_img, cutoff1)
            p99 = np.percentile(_img, cutoff2)
        else:
            p1 = np.percentile(img, cutoff1)
            p99 = np.percentile(img, cutoff2)

        img = skimage.exposure.rescale_intensity(img, in_range=(p1, p99), out_range=(0, 1))
        return (img * 65535).astype("uint16")

    def read(self, series, c):
        self.metadata._reader.setSeries(self.metadata.active_series[series])
        index = self.metadata._reader.getIndex(0, c, 0)
        byte_array = self.metadata._reader.openBytes(index)
        dtype = self.metadata.pixel_dtype  # no check required here as dtype has to be the same for all
        shape = self.metadata.tile_size(series)
        img = np.frombuffer(byte_array.tostring(), dtype=dtype).reshape(shape)

        if not self.do_rescale:
            return img
        else:
            if c not in self.no_rescale_channel:
                # get rescale_range for channel c
                if isinstance(self.rescale_range, dict):
                    rescale_range = self.rescale_range[c]
                else:
                    rescale_range = self.rescale_range

                # actually read the image with rescaling applied
                return self.rescale(img, rescale_range=rescale_range)
            else:
                return img



import io
import os
import platform
import shutil
from collections.abc import Callable
from contextlib import redirect_stdout
from functools import partial as func_partial
from pathlib import PosixPath

import h5py
import numpy as np
import pandas as pd
import pytorch_lightning as pl
import torch
from alphabase.io import tempmmap
from anndata import AnnData
from spatialdata.models import TableModel
from torchvision import transforms

from scportrait.pipeline._base import ProcessingStep
from scportrait.tools.ml.datasets import HDF5SingleCellDataset
from scportrait.tools.ml.plmodels import MultilabelSupervisedModel


class _FeaturizationBase(ProcessingStep):
    DEFAULT_DATA_LOADER = HDF5SingleCellDataset
    DEFAULT_MODEL_CLASS = MultilabelSupervisedModel
    PRETRAINED_MODEL_NAMES = [
        "autophagy_classifier",
    ]
    MASK_NAMES = ["nucleus", "cytosol"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._check_config()

        self.label = self.config["label"]
        self.num_workers = self.config["dataloader_worker_number"]
        self.batch_size = self.config["batch_size"]

        self.channel_names = None
        self.column_names = None
        self.dataset_size = None
        self.channel_selection = None
        self.inference_device = None
        self.model_class = None
        self.model = None
        self.transforms = None
        self.expected_imagesize = None
        self.data_type = None

        self._setup_channel_selection()

        # setup deep debugging
        self.deep_debug = False

        if "overwrite_run_path" not in self.__dict__.keys():
            self.overwrite_run_path = self.overwrite

    def _check_config(self) -> None:
        """Check if all required parameters are present in the config file."""

        assert "label" in self.config.keys(), "No label specified in config file."
        assert "dataloader_worker_number" in self.config.keys(), "No dataloader_worker_number specified in config file."
        assert "batch_size" in self.config.keys(), "No batch_size specified in config file."
        assert "inference_device" in self.config.keys(), "No inference_device specified in config file."

    def _setup_output(self) -> None:
        """Helper function to generate the output directory for the featurization results."""

        if not os.path.isdir(self.directory):
            os.makedirs(self.directory)

        if self.data_type is None:
            self.run_path = os.path.join(self.directory, self.label)
        else:
            self.run_path = os.path.join(self.directory, f"{self.data_type}_{self.label}")

        if not os.path.isdir(self.run_path):
            os.makedirs(self.run_path)
            self.log(f"Created new directory for featurization results: {self.run_path}")
        else:
            if self.overwrite:
                self.log("Overwrite flag is set, deleting existing directory for featurization results.")
                shutil.rmtree(self.run_path)
                os.makedirs(self.run_path)
                self.log(f"Created new directory for featurization results: {self.run_path}")
            elif self.overwrite_run_path:
                self.log("Overwrite flag is set, deleting existing directory for featurization results.")
                shutil.rmtree(self.run_path)
                os.makedirs(self.run_path)
                self.log(f"Created new directory for featurization results: {self.run_path}")
            else:
                raise ValueError(
                    f"Directory for featurization results already exists at {self.run_path}. Please set the overwrite flag to True if you wish to overwrite the existing directory."
                )

    def _setup_log_transform(self) -> None:
        """Setup if log transformation should be applied to the inference results."""
        if "log_transform" in self.config.keys():
            self.log_transform = self.config["log_transform"]
        else:
            self.log_transform = False  # default value

    def _setup_channel_selection(self) -> None:
        """Setup which channels should be used for inference. Default is that all channels available are used."""
        if "channel_selection" in self.config.keys():
            channel_selection = self.config["channel_selection"]
            if isinstance(channel_selection, list):
                assert all(
                    isinstance(x, int) for x in channel_selection
                ), "channel_selection should be a list of integers"
                self.channel_selection = channel_selection

            elif isinstance(channel_selection, int):
                self.channel_selection = [channel_selection]
            else:
                raise ValueError("channel_selection should be an integer or a list of integers.")

        else:
            self.channel_selection = None  # default value

    def _detect_automatic_inference_device(self) -> str:
        """Automatically detect the best inference device available on the system."""

        if torch.cuda.is_available():
            inference_device = "cuda"
        if torch.backends.mps.is_available():
            inference_device = torch.device("mps")
        else:
            inference_device = "cpu"

        return inference_device

    def _get_nmasks(self) -> None:
        """Get Number of segmentation Masks from the HDF5 file."""
        if "n_masks" not in self.__dict__.keys():
            if isinstance(self.extraction_file, str | PosixPath):
                try:
                    with h5py.File(self.extraction_file, "r") as f:
                        self.n_masks = f["n_masks"][()].item()
                except Exception as e:
                    raise ValueError(f"Could not extract number of masks from HDF5 file. Error: {e}") from e

            if isinstance(self.extraction_file, list):
                try:
                    n_masks = []
                    for file in self.extraction_file:
                        with h5py.File(file, "r") as f:
                            n_masks.append(f["n_masks"][()].item())
                    assert (
                        x == n_masks[0] for x in n_masks
                    ), "number of masks are not consistent over all passed HDF5 files."
                    self.n_masks = n_masks[0]
                except Exception as e:
                    raise ValueError(f"Could not extract number of masks from HDF5 file. Error: {e}") from e

    def _setup_inference_device(self) -> None:
        """
        Configure the featurization run to use the specified inference device.
        If no device is specified, the device is automatically detected.
        """

        if "inference_device" in self.config.keys():
            self.inference_device = self.config["inference_device"]

            # check that the selected inference device is also available
            if self.inference_device in ("cuda", torch.device("cuda")):
                if not torch.cuda.is_available():
                    if torch.backends.mps.is_available():
                        self.log(
                            "CUDA specified in config file but CUDA not available on system, switching to MPS for inference."
                        )
                        self.inference_device = torch.device("mps")
                    else:
                        self.log(
                            "CUDA specified in config file but CUDA not available on system, switching to CPU for inference."
                        )
                        self.inference_device = "cpu"
                else:
                    self.inference_device = torch.device("cuda")  # ensure that the complete device is always specified

            elif self.inference_device in ("mps", torch.device("mps")):
                if not torch.backends.mps.is_available():
                    if torch.cuda.is_available():
                        self.log(
                            "MPS specified in config file but MPS not available on system, switching to CUDA for inference."
                        )
                        self.inference_device = "cuda"
                    else:
                        self.log(
                            "MPS specified in config file but MPIS not available on system, switching to CPU for inference."
                        )
                        self.inference_device = "cpu"
                else:
                    self.inference_device = torch.device("mps")  # ensure that the complete device is always specified

            elif self.inference_device in ("automatic", "auto"):
                self.inference_device = self._detect_automatic_inference_device()
                self.log(f"Automatically configured inference device to {self.inference_device}")

            elif self.inference_device == "cpu":
                if torch.backends.mps.is_available():
                    self.log(
                        "CPU specified in config file but MPS available on system. Consider changing the device for the next run."
                    )
                if torch.cuda.is_available():
                    self.log(
                        "CPU specified in config file but CUDA available on system. Consider changing the device for the next run."
                    )
            else:
                raise ValueError(
                    "Invalid inference device specified in config file. Please use one of ['cuda', 'mps', 'cpu', 'automatic', 'auto']."
                )

        else:
            self.inference_device = self._detect_automatic_inference_device()
            self.log(f"Automatically configured inference device to {self.inference_device}")

    def _general_setup(self, dataset_paths: str | list[str], return_results: bool = False) -> None:
        """Helper function to execute all setup functions that are common to all featurization steps.

        Args:
            dataset_paths: Path to the extraction file or a list of paths.
            return_results: If True, the results are returned instead of being written to file

        Returns:
            None
        """

        self.extraction_file = dataset_paths
        if not return_results:
            self._setup_output()
        self._get_nmasks()
        self._setup_log_transform()
        self._setup_inference_device()

    def _get_model_specs(self) -> None:
        """Get the model"""
        # model location
        self.network_dir = self.config["network"]

        # get hyperparameters for loading model
        if "hparams_path" in self.config.keys():
            self.hparams_path = self.config["hparams_path"]
        else:
            self.hparams_path = None

        # model loading strategy: how to select which checkpoint to load if not a specific checkpoint is specified
        if "model_loading_strategy" in self.config.keys():
            strategy = self.config["model_loading_strategy"]
            if strategy not in ("max", "min", "latest", "path"):
                raise ValueError(
                    f"Invalid model loading strategy {strategy} specified. Please use one of ['max', 'min', 'latest', 'path']"
                )

            self.model_loading_strategy = self.config["model_loading_strategy"]
        else:
            self.model_loading_strategy = "max"  # default behvaiour is that the checkpoint with the highest epoch is used, in general it is highly recommended though to pass the path to a specific checkpoint file instead

        # Initiate the pytorch Lightning model class to which the checkpoing should be loaded
        if self.model_class is None:
            if "model_class" in self.config.keys():
                self.define_model_class(eval(self.config["model_class"]))
            else:
                self.define_model_class(self.DEFAULT_MODEL_CLASS)  # default model class
        else:
            self.log(
                f"Model class already defined as {self.model_class} will not overwrite. If this behaviour was unintended please set the model class to none by executing 'project.featurization_f.model_class = None'"
            )

        if "model_type" in self.config.keys():
            self.model_type = self.config["model_type"]
        else:
            self.model_type = None

    def _get_gpu_memory_usage(self):
        """Print the current memory usage on the GPU."""
        if self.inference_device == "cpu":
            return None

        elif self.inference_device == torch.device("cuda"):
            try:
                memory_usage = []

                for i in range(torch.cuda.device_count()):
                    gpu_memory = torch.cuda.memory_reserved(i) / 1024**2  # Convert bytes to MiB
                    memory_usage.append(gpu_memory)
                results = {f"GPU_{i}": f"{memory_usage[i]} MiB" for i in range(len(memory_usage))}
                return results
            except (RuntimeError, ValueError) as e:
                print("Error:", e)
                return None

        elif self.inference_device == torch.device("mps"):
            try:
                used_memory = torch.mps.driver_allocated_memory() + torch.mps.driver_allocated_memory()
                used_memory = used_memory / 1024**2  # Convert bytes to MiB
                return {"MPS": f"{used_memory} MiB"}
            except (RuntimeError, ValueError) as e:
                print("Error:", e)
                return None

        else:
            raise ValueError("Invalid inference device specified.")

    ### Functions for model loading and setup

    def _assign_model(self, model) -> None:
        """Save the model to the featurization object."""
        self.model = model

        # check if the hparams specify an expected image size
        if "hparams" in model.__dict__.keys():
            if "expected_imagesize" in model.hparams.keys():
                self.expected_imagesize = model.hparams["expected_imagesize"]

    def define_model_class(self, model_class, force_load=False) -> None:
        if isinstance(model_class, str):
            model_class = eval(model_class)  # convert string to class by evaluating it

        # check that it is a valid model class
        if force_load:
            if not issubclass(model_class, pl.LightningModule):
                Warning(
                    "Forcing the loading of the model class despite the provided model class not being a subclass of pl.LightningModule. This can lead to unexpected behaviour."
                )
            else:
                Warning(
                    "Forcing the loading of the model class is on but the provided model class is a subclass of pl.LightningModule. Consider setting the force_load parameter to False as this is not a recommended default setup."
                )
        else:
            if not issubclass(model_class, pl.LightningModule):
                raise ValueError(
                    "The provided model class is not a subclass of pl.LightningModule. Please provide a valid model class. If you are sure you wish to proceed with this, please reload the model class with the force_load parameter set to True."
                )

        self.model_class = model_class
        self.log(f"Model class defined as {model_class}")

    def _load_pretrained_model(self, model_name: str) -> pl.LightningModule:
        """
        Load a pretrained model from the SPARCScore library.

        Args:
            model_name : Name of the pretrained model to load.

        Returns:
            The loaded model.
        """

        if model_name == "autophagy_classifier":
            from scportrait.tools.ml.pretrained_models import autophagy_classifier

            model = autophagy_classifier(device=self.config["inference_device"])
            self.expected_imagesize = (128, 128)
        else:
            raise ValueError(
                f"Invalid model name {model_name} specified for pretrained model. The available pretrained models are {self.PRETRAINED_MODEL_NAMES}"
            )

        # set to eval mode and move to device
        model.eval()
        model.to(self.inference_device)

        return model

    def _load_model(
        self,
        ckpt_path,
        hparams_path: str | None = None,
        model_type: str | None = None,
    ) -> pl.LightningModule:
        """Load a model from a checkpoint file and transfer it to the inference device.

        Args:
            ckpt_path: Path to the checkpoint file.
            hparams_path: Path to the hparams file. If not provided, the hparams file is assumed to be in the same directory as the checkpoint file.
            model_type: Type of the model architecture to load. Default is None. For MultiLabelSupervisedModel, this can also be specified in the hparams file under the key model_type.

        Returns:
            The loaded model.
        """

        if self.model_class is None:
            raise ValueError("Model class not defined. Please define a model class before loading a model.")

        self.log(f"Loading model from checkpoint path: {ckpt_path}")

        # get path to hparams file

        if hparams_path is None:
            if self.debug:
                self.log(
                    "No hparams file provided, assuming hparams file is in the same directory as the checkpoint file. Trying to load from default location."
                )
            hparams_path = ckpt_path.replace(os.path.basename(ckpt_path), "hparams.yml")

            if not os.path.isfile(hparams_path):
                hparams_path = ckpt_path.replace(os.path.basename(ckpt_path), "hparams.yaml")

                if not os.path.isfile(hparams_path):
                    raise ValueError(
                        f"No hparams file found at {hparams_path}. Please provide a valid hparams file path."
                    )

            if self.debug:
                self.log(f"Loading hparams file from {hparams_path}")

        if model_type is None:
            model = self.model_class.load_from_checkpoint(
                checkpoint_path=ckpt_path,
                hparams_file=hparams_path,
                map_location=self.inference_device,
            )
        else:
            model = self.model_class.load_from_checkpoint(
                checkpoint_path=ckpt_path,
                hparams_file=hparams_path,
                map_location=self.inference_device,
                model_type=model_type,
            )

        model = model.eval()
        model.to(self.inference_device)

        if self.debug:
            self.log(f"model loaded, set to eval mode and transferred to device {self.inference_device}")

        return model

    def load_model(
        self,
        ckpt_path,
        hparams_path: str | None = None,
        model_type: str | None = None,
    ) -> None:
        model = self._load_model(ckpt_path, hparams_path, model_type)
        model.eval()
        self._assign_model(model)

    ### Functions regarding dataloading and transforms ####
    def configure_transforms(self, selected_transforms: list) -> None:
        self.transforms = transforms.Compose(selected_transforms)
        self.log(f"The following transforms were applied: {self.transforms}")

    def generate_dataloader(
        self,
        dataset_paths: str | list[str],
        dataset_labels: int | list[int] = 0,
        selected_transforms: transforms.Compose = transforms.Compose([]),
        size: int = 0,
        seed: int | None = 42,
        dataset_class=HDF5SingleCellDataset,
    ) -> torch.utils.data.DataLoader:
        """Create a pytorch dataloader from the provided single-cell image dataset.

        Args:
            dataset_paths: paths to the single-cell image datasets.
            selected_transforms:  List of transforms to apply to the images.
            size (optional):  Number of cells to select from the dataset. Default is 0, which means all samples are selected.
            seed (optional): Seed for the random number generator if splitting the dataset and only using a subset. Default is 42.

        Returns:
            The generated dataloader.

        """
        # generate dataset
        self.log(f"Reading data from path: {dataset_paths}")

        assert isinstance(
            self.transforms, transforms.Compose
        ), f"Transforms should be a torchvision.transforms.Compose object but recieved {self.transforms.__class__} instead."
        t = self.transforms

        if self.expected_imagesize is not None:
            self.log(f"Expected image size is set to {self.expected_imagesize}. Resizing images to this size.")
            t = transforms.Compose([t, transforms.Resize(self.expected_imagesize)])

        if isinstance(dataset_paths, list):
            assert isinstance(
                dataset_labels, list
            ), "If multiple directories are provided, multiple labels must be provided."
            paths = dataset_paths
            dataset_labels = dataset_labels
        elif isinstance(dataset_paths, str):
            assert isinstance(
                dataset_labels, int
            ), "If only one directory is provided, only one label must be provided."
            paths = [dataset_paths]
            dataset_labels = [dataset_labels]

        f = io.StringIO()
        with redirect_stdout(f):
            dataset = dataset_class(
                dir_list=paths,
                dir_labels=dataset_labels,
                transform=t,
                return_id=True,
                select_channel=self.channel_selection,
            )

        if size > 0:
            if size > len(dataset):
                raise ValueError(
                    f"Selected size {size} is larger than the dataset size {len(dataset)}. Please select a smaller size."
                )
            if size < len(dataset):
                residual_size = len(dataset) - size
                if seed is not None:
                    self.log(f"Using a seeded generator with seed {seed} to split dataset")
                    gen = torch.Generator()
                    gen.manual_seed(seed)
                    dataset, _ = torch.utils.data.random_split(dataset, [size, residual_size], generator=gen)
                else:
                    self.log("Using a random generator to split dataset.")
                    dataset, _ = torch.utils.data.random_split(dataset, [size, residual_size])

            # randomly select n elements from the dataset to process
            dataset = torch.utils.data.Subset(dataset, range(size))

        # save length of dataset for reaccess during inference
        self.dataset_size = len(dataset)
        self.log(f"Processing dataset with {self.dataset_size} cells")

        # check operating system
        if platform.system() == "Windows":
            context = "spawn"
            num_workers = 0  # need to disable multiprocessing otherwise it will throw an error
        elif platform.system() == "Darwin":
            context = "fork"
            num_workers = self.num_workers
        elif platform.system() == "Linux":
            context = "fork"
            num_workers = self.num_workers

        # generate dataloader

        if num_workers > 0:
            dataloader = torch.utils.data.DataLoader(
                dataset,
                batch_size=self.batch_size,
                shuffle=False,
                num_workers=num_workers,
                drop_last=False,
                multiprocessing_context=context,
            )
        else:
            dataloader = torch.utils.data.DataLoader(
                dataset,
                batch_size=self.batch_size,
                shuffle=False,
                num_workers=num_workers,
                drop_last=False,
            )

        self.log(
            f"Dataloader generated with a batchsize of {self.batch_size} and {self.num_workers} workers. Dataloader contains {len(dataloader)} entries."
        )

        return dataloader

    #### Inference functions ####
    def inference(
        self,
        dataloader: torch.utils.data.DataLoader,
        model_fun: Callable,
        pooler_output: bool = False,
        column_names: list | None = None,
        out_of_memory: bool = True,
    ) -> pd.DataFrame:
        """performs inference on a specific provided model and dataloader.

        Args:
            dataloader: Dataloader containing the data to perform inference on.
            model_fun: Model function to use for inference.
            pooler_output: If True, the args are passed as a ** call and the pooler output is returned. Defaults to False.
            column_names: Column names for the results dataframe. Defaults to None.


        Returns:
            pd.DataFrame: Dataframe containing the results of the inference.
        """
        self.log(f"Started processing of {len(dataloader)} batches.")

        data_iter = iter(dataloader)
        with torch.no_grad():
            # create id to track which index positions have already been filled in the results container
            ix: int = 0

            # perform first pass to get size of the returned inference results
            x, label, class_id = next(data_iter)
            if pooler_output:
                result = model_fun(**x.to(self.inference_device)).pooler_output.cpu().detach()
            else:
                result = model_fun(x.to(self.inference_device)).cpu().detach()

            # initialize a datastructure for saving the results
            n_entries, n_features = result.shape
            shape_features = (self.dataset_size, n_features)
            shape_labels = (self.dataset_size, 1)

            if out_of_memory:
                # use memory-mapped temp arrays to provide out-of-memory support
                features_path = tempmmap.create_empty_mmap(shape_features, dtype=np.float32)
                cell_ids_path = tempmmap.create_empty_mmap(shape_labels, dtype=np.int64)
                labels_path = tempmmap.create_empty_mmap(shape_labels, dtype=np.int64)

                features = tempmmap.mmap_array_from_path(features_path)
                cell_ids = tempmmap.mmap_array_from_path(cell_ids_path)
                labels = tempmmap.mmap_array_from_path(labels_path)

            else:
                # use numpy arrays
                features = np.zeros(shape_features, dtype=np.float32)
                cell_ids = np.zeros(shape_labels, dtype=np.int64)
                labels = np.zeros(shape_labels, dtype=np.int64)

            # save the results for each batch into the storage container at the specified indices
            features[ix : (ix + result.shape[0])] = result.numpy()
            cell_ids[ix : (ix + result.shape[0])] = class_id.unsqueeze(1)
            labels[ix : (ix + result.shape[0])] = label.unsqueeze(1)
            ix += result.shape[0]  # update id to track filled positions

            # add check to ensure this only runs if we have more than one batch in the dataset
            if len(dataloader) > 1:
                for i in range(len(dataloader) - 1):
                    if i % 10 == 0:
                        self.log(f"processing batch {i}")

                    x, label, class_id = next(data_iter)
                    if pooler_output:
                        result = model_fun(**x.to(self.inference_device)).pooler_output.cpu().detach()
                    else:
                        result = model_fun(x.to(self.inference_device)).cpu().detach()

                    # save the results for each batch into the storage container at the specified indices
                    features[ix : (ix + result.shape[0])] = result.numpy()
                    cell_ids[ix : (ix + result.shape[0])] = class_id.unsqueeze(1)
                    labels[ix : (ix + result.shape[0])] = label.unsqueeze(1)
                    ix += result.shape[0]  # update id to track filled positions

        if self.log_transform:
            self.log("Applying log transformation to results.")
            sigma = 1e-9  # to avoid log(0)
            features = np.log(features + sigma)

        # save inferred activations / predictions
        if column_names is None:
            column_names = [f"result_{i}" for i in range(result.shape[1])]

        dataframe = pd.DataFrame(data=features, columns=column_names)
        dataframe["label"] = labels
        dataframe["cell_id"] = cell_ids.astype("int")

        self.log("finished processing.")

        return dataframe

    #### Results writing functions ####

    def _write_results_csv(self, results: pd.DataFrame, path: str | PosixPath) -> None:
        """Write results to a CSV file."""
        results.to_csv(path, index=False)
        self.log(f"Results saved to file: {path}")

    def _write_results_sdata(self, results: pd.DataFrame, label: str, mask_type: str = "seg_all") -> None:
        """Add results to the spatialdata object.

        Args:
            results: Results to add to the spatialdata object.
            label: Label for the results.
            mask_type: Type of mask used for the results. Defaults to "seg_all".
        """
        results.set_index("cell_id", inplace=True)
        results.drop(columns=["label"], inplace=True)

        feature_matrix = results.to_numpy()
        var_names = results.columns
        obs_indices = results.index.astype(str)

        if self.project.nuc_seg_status:
            # save nucleus segmentation
            obs = pd.DataFrame()
            obs.index = obs_indices
            obs["instance_id"] = obs_indices.astype(
                int
            )  # this needs to be int otherwise data will not be loaded correctly
            obs["region"] = f"{mask_type}_{self.MASK_NAMES[0]}"
            obs["region"] = obs["region"].astype("category")

            table = AnnData(X=feature_matrix, var=pd.DataFrame(index=var_names), obs=obs)
            table = TableModel.parse(
                table,
                region=[f"{mask_type}_{self.MASK_NAMES[0]}"],
                region_key="region",
                instance_key="instance_id",
            )

            self.filehandler._write_table_object_sdata(
                table,
                f"{self.__class__.__name__ }_{label}_{self.MASK_NAMES[0]}",
                overwrite=self.overwrite_run_path,
            )

        if self.project.cyto_seg_status:
            # save cytoplasm segmentation
            obs = pd.DataFrame()
            obs.index = obs_indices
            obs["instance_id"] = obs_indices.astype(
                int
            )  # this needs to be int otherwise data will not be loaded correctly
            obs["region"] = f"{mask_type}_{self.MASK_NAMES[1]}"
            obs["region"] = obs["region"].astype("category")

            table = AnnData(X=feature_matrix, var=pd.DataFrame(index=var_names), obs=obs)
            table = TableModel.parse(
                table,
                region=[f"{mask_type}_{self.MASK_NAMES[1]}"],
                region_key="region",
                instance_key="instance_id",
            )

            self.filehandler._write_table_object_sdata(
                table,
                f"{self.__class__.__name__ }_{label}_{self.MASK_NAMES[1]}",
                overwrite=self.overwrite_run_path,
            )

    #### Cleanup Functions ####

    def _post_processing_cleanup(self) -> None:
        """reset all attribute values to the default parameters."""
        if self.debug:
            memory_usage = self._get_gpu_memory_usage()
            self.log(f"GPU memory before performing cleanup: {memory_usage}")

        if "dataloader" in self.__dict__.keys():
            del self.dataloader  # type: ignore

        if "models" in self.__dict__.keys():
            del self.models  # type: ignore

        if "model" in self.__dict__.keys():
            del self.model  # type: ignore

        if "overwrite_run_path" in self.__dict__.keys():
            del self.overwrite_run_path  # type: ignore

        if "n_masks" in self.__dict__.keys():
            del self.n_masks  # type: ignore

        if "data_type" in self.__dict__.keys():
            del self.data_type  # type: ignore

        # reset to init values to ensure that subsequent runs are not affected by previous runs
        self.log_transform = None
        self.channel_names = None
        self.column_names = None
        self.dataset_size = None
        self.model_class = None
        self.transforms = None
        self.channel_selection = None
        self.model = None

        self._clear_cache()

        if self.debug:
            memory_usage = self._get_gpu_memory_usage()
            self.log(f"GPU memory after performing cleanup: {memory_usage}")

        # this needs to be called after the memory usage has been assesed
        if "inference_device" in self.__dict__.keys():
            del self.inference_device


###############################################
###### DeepLearning based Classification ######
###############################################


class MLClusterClassifier(_FeaturizationBase):
    """
    Class for classifying single cells using a pre-trained machine learning model.

    This class takes a pre-trained model and uses it to classify single cells,
    using the model's forward function or encoder function, depending on the
    user's choice. The classification results are saved to a CSV file.
    """

    CLEAN_LOG = True
    DEFAULT_LOG_NAME = "processing_MLClusterClassifier.log"

    def __init__(self, *args, **kwargs):
        """
        Class is initiated to classify extracted single cells.

        Args:
            config : Configuration for the extraction passed over from the :class:`pipeline.Project`.
            directory: Directory for the extraction log and results. Will be created if not existing yet.
            debug : Flag used to output debug information and map images.
            overwrite : Flag used to overwrite existing results.
        """
        super().__init__(*args, **kwargs)

        if self.CLEAN_LOG:
            self._clean_log_file()

    def _get_network_dir(self) -> pl.LightningModule:
        if self.network_dir in self.PRETRAINED_MODEL_NAMES:
            pass
        else:
            if self.model_loading_strategy == "path":
                pass

            elif self.network_dir.endswith(".ckpt"):
                Warning(
                    "Provided network ends in .ckpt, assuming this is a complete model path. To avoid this warning in the future please set the config parameter 'model_loading_strategy' to 'path'."
                )
                self.log(
                    "Provided network ends in .ckpt, assuming this is a complete model path. To avoid this warning in the future please set the config parameter 'model_loading_strategy' to 'path'."
                )
            else:
                # then we assume this is a directory containing multiple checkpoints and we want to load a specific one based on the model_loading_strategy

                model_path = self.network_dir
                checkpoints = [file for file in os.listdir(model_path) if file.endswith(".ckpt")]
                checkpoints.sort()

                if len(checkpoints) < 1:
                    raise ValueError(f"No checkpoint files found at {model_path}")

                if len(checkpoints) == 1:
                    self.log("Only one checkpoint found in network directory.")

                    # update network_dir to the path of the checkpoint
                    self.network_dir = os.path.join(model_path, checkpoints[0])
                else:
                    if self.model_loading_strategy == "max":
                        max_epoch = max([int(file.split("epoch=")[1].split("-")[0]) for file in checkpoints])
                        max_epoch_file = [file for file in checkpoints if f"epoch={max_epoch}" in file][0]
                        self.network_dir = os.path.join(model_path, max_epoch_file)
                        self.log(
                            f"Using model selection strategy 'max', selecting model with the maximum epoch {max_epoch} from path {self.network_dir}"
                        )
                    elif self.model_loading_strategy == "min":
                        min_epoch = min([int(file.split("epoch=")[1].split("-")[0]) for file in checkpoints])
                        min_epoch_file = [file for file in checkpoints if f"epoch={min_epoch}" in file][0]
                        self.network_dir = os.path.join(model_path, min_epoch_file)
                        self.log(
                            f"Using model selection strategy 'min', selecting model with the minimum epoch {min_epoch} from path {self.network_dir}"
                        )
                    elif self.model_loading_strategy == "latest":
                        self.network_dir = os.path.join(model_path, checkpoints[-1])
                        self.log(
                            f"Using model selection strategy 'latest', selecting the latest model from path {self.network_dir}"
                        )
                    else:
                        raise ValueError(
                            f"Invalid model loading strategy {self.model_loading_strategy} specified. Please use one of ['max', 'min', 'latest'] if not provding a path to a model cpkt."
                        )

    def _setup_encoders(self) -> None:
        # extract which inferences to make from config file
        if "encoders" in self.config.keys():
            encoders = self.config["encoders"]
        else:
            encoders = ["forward"]  # default parameter

        self.models = []

        for encoder in encoders:
            if encoder == "forward":
                self.models.append(self.model.network.forward)
            if encoder == "encoder":
                self.models.append(self.model.network.encoder)

    def _setup_transforms(self) -> None:
        if self.transforms is not None:
            self.log(
                "Transforms already configured manually. Will not overwrite. If this behaviour was unintended please set the transforms to None by executing 'project.featurization_f.transforms = None'"
            )

        if "transforms" in self.config.keys():
            self.transforms = eval(self.config["transforms"])
        else:
            self.transforms = transforms.Compose([])  # default is no transforms

    def _setup(self, dataset_paths: str, return_results: bool) -> None:
        self._general_setup(dataset_paths=dataset_paths, return_results=return_results)
        self._get_model_specs()
        self._get_network_dir()

        if self.network_dir in self.PRETRAINED_MODEL_NAMES:
            model = self._load_pretrained_model(self.network_dir)
        else:
            model = self._load_model(
                ckpt_path=self.network_dir,
                hparams_path=self.hparams_path,
                model_type=self.model_type,
            )

        self._assign_model(model)

        self._setup_encoders()
        self._setup_transforms()

    def process(
        self,
        dataset_paths: str,
        dataset_labels: int | list[int] = 0,
        size: int = 0,
        return_results: bool = False,
    ) -> None | list[pd.DataFrame]:
        """
        Perform classification on the provided HDF5 dataset.

        Args:
            dataset_paths : Directory containing the extracted HDF5 files from the project. If this class is used as part of a project processing workflow, this argument will be provided automatically.
            size : How many cells should be selected for inference. Default is 0, which means all cells are selected.

        Returns:
            None unless return_results is True, the results are returned as a list of pandas DataFrames. Otherwise, the results are written to CSV files located in the project directory.

        Important:
            If this class is used as part of a project processing workflow, the first argument will be provided by the
            `Project` class based on the previous single-cell extraction. Therefore, only the second and third arguments
            need to be provided. The `Project` class will automatically provide the most recent extracted single-cell
            dataset together with the supplied parameters.

        Examples:
            ```python
            project.classify()
            ```

        Notes:
            The following parameters are required in the config file:

            ```yaml
            MLClusterClassifier:
                # Channel number on which the classification should be performed
                channel_selection: 4

                # Number of threads to use for dataloader
                dataloader_worker_number: 24

                # Batch size to pass to GPU
                batch_size: 900

                # Path to PyTorch checkpoint that should be used for inference
                network: "path/to/model/"

                # Classifier architecture implemented in scPortrait
                # Choose one of VGG1, VGG2, VGG1_old, VGG2_old
                classifier_architecture: "VGG2_old"

                # If more than one checkpoint is provided in the network directory, which checkpoint should be chosen
                # Should either be "max" or a numeric value indicating the epoch number
                epoch: "max"

                # Name of the classifier used for saving the classification results to a directory
                label: "Autophagy_15h_classifier1"

                # List of which inference methods should be performed
                # Available: "forward" and "encoder"
                # If "forward": images are passed through all layers of the model and the final inference results are written to file
                # If "encoder": activations at the end of the CNN are written to file
                encoders: ["forward", "encoder"]

                # On which device inference should be performed
                # For speed, should be "cuda"
                inference_device: "cuda"

                #define dataset transforms
                transforms:
                    resize: 128
            ```
        """
        self.log("Started MLClusterClassifier classification.")

        # perform setup
        self._setup(dataset_paths=dataset_paths, return_results=return_results)

        self.dataloader = self.generate_dataloader(
            dataset_paths,
            dataset_labels=dataset_labels,
            selected_transforms=self.transforms,
            size=size,
            dataset_class=self.DEFAULT_DATA_LOADER,
        )

        # perform inference
        all_results = []
        for model in self.models:
            self.log(f"Starting inference for model encoder {model.__name__}")
            results = self.inference(self.dataloader, model)

            if not return_results:
                output_name = f"inference_{model.__name__}"
                path = os.path.join(self.run_path, f"{output_name}.csv")

                self._write_results_csv(results, path)
                self._write_results_sdata(results, label=f"{self.label}_{model.__name__}")
            else:
                all_results.append(results)

        if return_results:
            self._clear_cache()
            return all_results
        else:
            self.log(f"Results saved to file: {path}")
            # perform post processing cleanup
            if not self.deep_debug:
                self._post_processing_cleanup()
            return None


class EnsembleClassifier(_FeaturizationBase):
    """
    This class takes a pre-trained ensemble of models and uses it to classify extracted single cell datasets.
    """

    CLEAN_LOG = True
    DEFAULT_LOG_NAME = "processing_EnsembleClassifier.log"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.CLEAN_LOG:
            self._clean_log_file()

    def _setup_transforms(self):
        if self.transforms is not None:
            self.log(
                "Transforms already configured manually. Will not overwrite. If this behaviour was unintended please set the transforms to None by executing 'project.classification_f.transforms = None'"
            )
            return
        else:
            self.transforms = transforms.Compose([])

    def _load_models(self):
        # load models and generate ensemble
        self.model = []
        self.model_names = []

        for model_name, model_path in self.network_dir.items():
            model = self._load_model(ckpt_path=model_path, hparams_path=self.hparams_path)

            # check for hparams expected_imagesize
            if self.expected_imagesize is None:
                if "expected_imagesize" in model.hparams.keys():
                    self.expected_imagesize = model.hparams["expected_imagesize"]
            else:
                if "expected_imagesize" in model.hparams.keys():
                    if self.expected_imagesize != model.hparams["expected_imagesize"]:
                        raise ValueError("Expected image sizes of models in ensemble do not match.")

            self.model.append(model)
            self.model_names.append(model_name)

        self.log(f"Model Ensemble generated with a total of {len(self.model)} models.")
        memory_usage = self._get_gpu_memory_usage()
        self.log(f"GPU memory usage after loading models: {memory_usage}")

    def _setup(self, dataset_paths: str, return_results: bool):
        self._general_setup(dataset_paths=dataset_paths, return_results=return_results)
        self._get_model_specs()
        self._setup_transforms()

        # ensure that the network_dir is a dictionary
        if not isinstance(self.network_dir, dict):
            raise ValueError(
                "network_dir should be a dictionary containing the model names and paths to the model checkpoints."
            )

        self._load_models()

    def process(
        self, dataset_paths: str, dataset_labels: int | list[int] = 0, size: int = 0, return_results: bool = False
    ) -> None | dict:
        """
        Function called to perform classification on the provided HDF5 dataset.

        Args:
            dataset_paths (str): Directory containing the extracted HDF5 files from the project. If this class is used as part of
            a project processing workflow this argument will be provided automatically.

        Returns:
            None: Results are written to csv files located in the project directory.

        Important:
            If this class is used as part of a project processing workflow, the first argument will be provided by the ``Project``
            class based on the previous single-cell extraction. Therefore, no parameters need to be provided

        Example:

            .. code-block:: python

                project.classify()

        Note:

            The following parameters are required in the config file:

            .. code-block:: yaml

                EnsembleClassifier:
                    # channel number on which the classification should be performed
                    channel_selection: 4

                    #number of threads to use for dataloader
                    dataloader_worker_number: 24

                    #batch size to pass to GPU
                    batch_size: 900

                    #path to pytorch checkpoint that should be used for inference
                    networks:
                        model1: "path/to/model1/"
                        model2: "path/to/model2/"

                    #specify input size that the models expect, provided images will be rescaled to this size
                    input_image_px: 128

                    #label under which the results will be saved
                    classification_label: "Autophagy_15h_classifier1"

                    # on which device inference should be performed
                    # for speed should be "cuda"
                    inference_device: "cuda"
        """

        self.log("Starting Ensemble Classification")

        self._setup(dataset_paths=dataset_paths, return_results=return_results)

        self.dataloader = self.generate_dataloader(
            dataset_paths,
            dataset_labels=dataset_labels,
            selected_transforms=self.transforms,
            size=size,
            dataset_class=self.DEFAULT_DATA_LOADER,
        )

        # perform inference
        all_results = {}
        for model_name, model in zip(self.model_names, self.model, strict=False):
            self.log(f"Starting inference for model {model_name}")
            results = self.inference(self.dataloader, model)

            output_name = f"ensemble_inference_{model_name}"

            if not return_results:
                path = os.path.join(self.run_path, f"{output_name}.csv")

                self._write_results_csv(results, path)
                self._write_results_sdata(results, label=model_name)
            else:
                all_results[model_name] = results

        if return_results:
            self._clear_cache()
            return all_results
        else:
            # perform post processing cleanup
            if not self.deep_debug:
                self._post_processing_cleanup()
            return None


class ConvNeXtFeaturizer(_FeaturizationBase):
    CLEAN_LOG = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.CLEAN_LOG:
            self._clean_log_file()

        self._check_config()

        # assert that the correct transformers version is installed
        try:
            import transformers
        except ImportError:
            raise ImportError(
                "transformers is not installed. Please install it via pip install 'transformers==4.26.0'"
            ) from None

        assert (
            transformers.__version__ == "4.26.0"
        ), "Please install transformers version 4.26.0 via pip install --force 'transformers==4.26.0'"

        assert len(self.channel_selection) in [1, 3], "channel_selection should be either 1 or 3 channels"

    def _load_model(self):
        # lazy imports
        from transformers import ConvNextModel

        # silence warnings from transformers that are not relevant here
        # we do actually just want to load some of the weights to access the convnext features

        model = ConvNextModel.from_pretrained("facebook/convnext-xlarge-224-22k")
        model.eval()
        model.to(self.inference_device)

        self._assign_model(model)

    def _silence_warnings(self):
        import logging

        from transformers import logging as hf_logging

        # Create a custom filter class to suppress specific warnings from huggingfaces transformers
        class SpecificMessageFilter(logging.Filter):
            def __init__(self, suppressed_keywords):
                super().__init__()
                self.suppressed_keywords = suppressed_keywords

            def filter(self, record):
                return not any(keyword in record.getMessage() for keyword in self.suppressed_keywords)

        # Keywords to suppress
        suppressed_keywords = [
            "Some weights of the model checkpoint at facebook",
            "Could not find image processor class in the image processor config",
        ]

        transformers_logger = hf_logging.get_logger()
        for handler in transformers_logger.handlers:
            handler.addFilter(SpecificMessageFilter(suppressed_keywords))

    def _setup_transforms(self) -> None:
        # lazy imports
        from transformers import AutoImageProcessor

        from scportrait.tools.ml.transforms import ChannelMultiplier

        feature_extractor = AutoImageProcessor.from_pretrained("facebook/convnext-xlarge-224-22k")

        # custom transform to properly pass images to model
        def get_pixel_values(in_tensor):
            in_tensor["pixel_values"] = in_tensor["pixel_values"][0]
            return in_tensor

        if len(self.channel_selection) == 1:
            self.transforms = transforms.Compose(
                [
                    ChannelMultiplier(3),
                    feature_extractor,
                    get_pixel_values,
                ]
            )
        elif len(self.channel_selection) == 3:
            self.transforms = transforms.Compose([feature_extractor, get_pixel_values])
        else:
            raise ValueError("channel_selection should be either 1 or 3 channels")

    def _generate_column_names(self) -> list:
        N_CONVNEXT_FEATURES = 2048
        column_names = [f"convnext_feature_{i}" for i in range(N_CONVNEXT_FEATURES)]
        return column_names

    def _setup(self, dataset_paths: str | list[str], return_results: bool) -> None:
        self._silence_warnings()
        self._general_setup(dataset_paths=dataset_paths, return_results=return_results)
        self._load_model()
        self._setup_transforms()
        self._load_model()

    def process(
        self,
        dataset_paths: str | list[str],
        dataset_labels: int | list[int] = 0,
        size: int = 0,
        return_results: bool = False,
    ) -> None | pd.DataFrame:
        """
        Perform ConvNeXt inference on the provided HDF5 dataset.

        Args
            dataset_paths : Paths to the single-cell HDF5 files on which inference should be performed. If this class is used as part of a project processing workflow this argument will be provided automatically.
            dataset_labels: labels for the provided single-cell image datasets
            size : How many cells should be selected for inference. Default is 0, meaning all cells are selected.
            return_results : If True, the results are returned as a pandas DataFrame. Otherwise the results are written out to file.

        Returns:
            None if return_results is False, otherwise a pandas DataFrame containing the results.

        Important
        ---------
        If this class is used as part of a project processing workflow, the first argument will be provided by the ``Project``
        class based on the previous single-cell extraction. Therefore, only the second and third arguments need to be provided.
        The Project class will automatically provide the most recent extracted single-cell dataset together with the supplied parameters.

        Examples
        --------
        .. code-block:: python
            project.classify()

        Notes
        -----
        The following parameters are required in the config file:

        .. code-block:: yaml

            MLClusterClassifier:
                # Channel index on which the classification should be performed
                channel_selection: 4

                # Number of threads to use for dataloader
                dataloader_worker_number: 24

                # Batch size to pass to GPU
                batch_size: 100

                # On which device inference should be performed
                # For speed, should be "cuda"
                inference_device: "cuda"
        """

        self._setup(dataset_paths=dataset_paths, return_results=return_results)

        self.dataloader = self.generate_dataloader(
            dataset_paths,
            dataset_labels=dataset_labels,
            selected_transforms=self.transforms,
            size=size,
            dataset_class=self.DEFAULT_DATA_LOADER,
        )

        results = self.inference(
            self.dataloader, self.model, pooler_output=True, column_names=self._generate_column_names()
        )

        if return_results:
            self._clear_cache()
            return results
        else:
            output_name = "calculated_image_features"
            path = os.path.join(self.run_path, f"{output_name}.csv")

            self._write_results_csv(results, path)
            self._write_results_sdata(results, label="ConvNeXt")

            # perform post processing cleanup
            if not self.deep_debug:
                self._post_processing_cleanup()
            return None


####### CellFeaturization based on Classic Featurecalculation #######
class _cellFeaturizerBase(_FeaturizationBase):
    CLEAN_LOG = True

    # define the output column names
    MASK_NAMES = ["nucleus", "cytosol", "cytosol_only"]
    MASK_STATISTICS = ["area"]
    CHANNEL_STATISTICS = [
        "mean",
        "median",
        "quant75",
        "quant25",
        "summed_intensity",
        "summed_intensity_area_normalized",
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.CLEAN_LOG:
            self._clean_log_file()

    def _setup_transforms(self):
        if self.transforms is not None:
            self.log(
                "Transforms already configured manually. Will not overwrite. If this behaviour was unintended please set the transforms to None by executing 'project.classification_f.transforms = None'"
            )
            return

        self.transforms = transforms.Compose([])
        return

    def _get_channel_specs(self):
        if self.project is None:
            if isinstance(self.extraction_file, str):
                with h5py.File(self.extraction_file, "r") as f:
                    self.channel_names = list(f["channel_information"][:].astype(str))
            if isinstance(self.extraction_file, list):
                channel_names = []
                for file in self.extraction_file:
                    with h5py.File(file, "r") as f:
                        channel_names.append(list(f["channel_information"][:].astype(str)))
                assert (
                    x == channel_names[0] for x in channel_names
                ), "Channel names are not consistent over all passed HDF5 files."
                self.channel_names = channel_names[0]
        else:
            if "channel_names" in self.project.__dict__.keys():
                self.channel_names = self.project.channel_names
            else:
                self.channel_names = self.project.input_image.c.values

    def _generate_column_names(
        self,
        n_masks: int = 2,
        n_channels: int = 3,
        channel_names: list | None = None,
    ) -> None:
        column_names = []

        if n_masks == 1:
            self.project.get_project_status()

            if self.project.nuc_seg_status:
                mask_name = self.MASK_NAMES[0]
            elif self.project.cyto_seg_status:
                mask_name = self.MASK_NAMES[1]
            else:
                raise ValueError("no segmentation mask found in sdata object.")
            mask_names = [mask_name]

        elif n_masks == 2:
            mask_names = self.MASK_NAMES

        # get the mask names with the mask attributes
        for mask in mask_names:
            for mask_stat in self.MASK_STATISTICS:
                column_names.append(f"{mask}_{mask_stat}")

        if channel_names is None:
            channel_names = [f"channel_{i}" for i in range(n_channels)]

        for channel_name in channel_names:
            for mask in mask_names:
                for channel_stat in self.CHANNEL_STATISTICS:
                    column_names.append(f"{channel_name}_{channel_stat}_{mask}")

        self.column_names = column_names

    def calculate_statistics(self, img: torch.Tensor, n_masks: int = 2):
        """
        Calculate statistics for an image batch.

        Args:
            img : Tensor containing the image batch.
            n_masks : Number of masks in the image. Masks are always the first images in the image stack. Default is 2.

        Returns:
            The calculated image statistics
        """
        N, _, _, _ = img.shape

        mask_statistics = []
        masks = []

        for i in range(n_masks):
            mask = img[:, i] > 0
            area = mask.view(N, -1).sum(1, keepdims=True)

            masks.append(mask)
            mask_statistics.append(area)

        if n_masks == 2:
            mask = masks[1] ^ masks[0]
            area = mask.view(N, -1).sum(1, keepdims=True)

            masks.append(mask)
            mask_statistics.append(area)

        channel_statistics = []

        for channel in range(n_masks, img.shape[1]):
            img_selected = img[:, channel]

            for mask in masks:
                mask[mask == 0] = torch.nan

                # apply mask to channel to only compute the statistics over the pixels that are relevant
                _img_selected = (img_selected * mask).to(
                    torch.float32
                )  # ensure we have correct dytpe for subsequent calculations

                mean = _img_selected.view(N, -1).nanmean(1, keepdim=True)
                median = _img_selected.view(N, -1).nanquantile(q=0.5, dim=1, keepdim=True)
                quant75 = _img_selected.view(N, -1).nanquantile(q=0.75, dim=1, keepdim=True)
                quant25 = _img_selected.view(N, -1).nanquantile(q=0.25, dim=1, keepdim=True)
                summed_intensity = _img_selected.view(N, -1).sum(1, keepdim=True)
                summed_intensity_area_normalized = summed_intensity / mask_statistics[-1]

                # save results
                channel_statistics.extend(
                    [
                        mean,
                        median,
                        quant75,
                        quant25,
                        summed_intensity,
                        summed_intensity_area_normalized,
                    ]
                )

        # Generate results tensor with all values and return
        items = mask_statistics + channel_statistics
        results = torch.concat(
            items,
            1,
        )

        return results

    def _write_results_sdata(self, results, mask_type="seg_all"):
        if self.project.nuc_seg_status:
            # save nucleus segmentation
            columns_drop = [x for x in results.columns if self.MASK_NAMES[1] in x]
            segmentation_name = f"{mask_type}_{self.MASK_NAMES[0]}"

            if self.channel_selection is not None:
                table_name = f"{self.__class__.__name__ }_{self.config['channel_selection']}_{self.MASK_NAMES[0]}"
            else:
                table_name = f"{self.__class__.__name__ }_{self.MASK_NAMES[0]}"

        if self.project.cyto_seg_status:
            # save cytosol segmentation
            columns_drop = [x for x in results.columns if self.MASK_NAMES[0] in x]
            segmentation_name = f"{mask_type}_{self.MASK_NAMES[1]}"

            # define name to save table under
            if self.channel_selection is not None:
                table_name = f"{self.__class__.__name__ }_{self.config['channel_selection']}_{self.MASK_NAMES[1]}"
            else:
                table_name = f"{self.__class__.__name__ }_{self.MASK_NAMES[1]}"

        _results = results.drop(columns=columns_drop)
        _results.set_index("cell_id", inplace=True)
        _results.drop(columns=["label"], inplace=True)

        feature_matrix = _results.to_numpy()
        var_names = _results.columns
        obs_indices = _results.index.astype(str)

        adata = AnnData(X=feature_matrix, var=pd.DataFrame(index=var_names), obs=pd.DataFrame(index=obs_indices))
        self.filehandler._write_table_sdata(
            adata, segmentation_mask_name=segmentation_name, table_name=table_name, overwrite=self.overwrite_run_path
        )


class CellFeaturizer(_cellFeaturizerBase):
    """
    Class for extracting general image features from SPARCS single-cell image datasets.
    The extracted features are saved to a CSV file. The features are calculated on the basis of a specified channel.

    The features which are calculated are:

    - Area of the masks in pixels
    - Mean intensity of the chosen channel in the regions labelled by each of the masks
    - Median intensity of the chosen channel in the regions labelled by each of the masks
    - 75% quantile of the chosen channel in the regions labelled by each of the masks
    - 25% quantile of the chosen channel in the regions labelled by each of the masks
    - Summed intensity of the chosen channel in the regions labelled by each of the masks
    - Summed intensity of the chosen channel in the region labelled by each of the masks normalized for area

    The features are outputed in this order in the CSV file.
    """

    DEFAULT_LOG_NAME = "processing_CellFeaturizer.log"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.channel_selection = None  # ensure that all images are passed to the function

    def _setup(self, dataset_paths: str | list[str], return_results: bool):
        self._general_setup(dataset_paths=dataset_paths, return_results=return_results)
        self._setup_transforms()
        self._get_channel_specs()

    def process(
        self,
        dataset_paths: str | list[str],
        dataset_labels: int | list[int] = 0,
        size: int = 0,
        return_results: bool = False,
    ) -> None | pd.DataFrame:
        """
        Perform featurization on the provided HDF5 dataset.

        Args:
            dataset_paths : Paths to the single-cell HDF5 files on which inference should be performed. If this class is used as part of a project processing workflow this argument will be provided automatically.
            dataset_labels: labels for the provided single-cell image datasets
            size : How many cells should be selected for inference. Default is 0, meaning all cells are selected.
            return_results : If True, the results are returned as a pandas DataFrame. Otherwise the results are written out to file.

        Returns:
            None if return_results is False, otherwise a pandas DataFrame containing the results.

        Important
        ---------
        If this class is used as part of a project processing workflow, the first argument will be provided by the ``Project`` class based on the previous single-cell extraction. Therefore, only the second and third argument need to be provided. The Project class will automatically provide the most recent extraction results together with the supplied parameters.

        Notes
        -----
        The following parameters are required in the config file:

        .. code-block:: yaml

            CellFeaturizer:
                # Channel number on which the featurization should be performed
                channel_selection: 4

                # Number of threads to use for dataloader
                dataloader_worker_number: 0 # needs to be 0 if using CPU

                # Batch size to pass to GPU
                batch_size: 900

                # On which device inference should be performed
                # For speed should be "cuda"
                inference_device: "cpu"

                # Label under which the results should be saved
                screen_label: "Ch3_Featurization"
        """
        self.log("Started CellFeaturization of all available channels.")

        # perform setup
        self._setup(dataset_paths=dataset_paths, return_results=return_results)

        self.dataloader = self.generate_dataloader(
            dataset_paths,
            dataset_labels=dataset_labels,
            selected_transforms=self.transforms,
            size=size,
            dataset_class=self.DEFAULT_DATA_LOADER,
        )

        # get first example image from dataloader
        x, _, _ = next(iter(self.dataloader))
        N, c, x, y = x.shape

        # perform sanity check on the number of channels
        assert (
            (self.n_masks + len(self.channel_names)) == c
        ), f"Number of images in the dataset ({c}) does not match the number of masks ({self.n_masks}) and channel names ({len(self.channel_names)}) specified in the project."

        # generate column names
        self._generate_column_names(n_masks=self.n_masks, n_channels=c, channel_names=self.channel_names)

        # define inference function
        f = func_partial(self.calculate_statistics, n_masks=self.n_masks)

        results = self.inference(
            self.dataloader,
            f,
            column_names=self.column_names,
        )

        if return_results:
            self._clear_cache()
            return results
        else:
            output_name = "calculated_image_features"
            path = os.path.join(self.run_path, f"{output_name}.csv")

            self._write_results_csv(results, path)
            self._write_results_sdata(results)

            # perform post processing cleanup
            if not self.deep_debug:
                self._post_processing_cleanup()
            self._clear_cache()
            return None


class CellFeaturizer_single_channel(_cellFeaturizerBase):
    DEFAULT_LOG_NAME = "processing_CellFeaturizer.log"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _setup_channel_selection(self):
        if self.n_masks == 2:
            self.channel_selection = [0, 1, self.channel_selection]
        if self.n_masks == 1:
            self.channel_selection = [0, self.channel_selection]
        return

    def _setup(self, dataset_paths: str | list[str], return_results: bool):
        self._general_setup(dataset_paths=dataset_paths, return_results=return_results)
        self._setup_channel_selection()
        self._setup_transforms()
        self._get_channel_specs()

    def process(
        self, dataset_paths: str | list[str], dataset_labels: int | list[int] = 0, size=0, return_results: bool = False
    ) -> None | pd.DataFrame:
        self.log(f"Started CellFeaturization of selected channel {self.channel_selection}.")

        # perform setup
        self._setup(dataset_paths=dataset_paths, return_results=return_results)

        self.dataloader = self.generate_dataloader(
            dataset_paths,
            dataset_labels=dataset_labels,
            selected_transforms=self.transforms,
            size=size,
            dataset_class=self.DEFAULT_DATA_LOADER,
        )

        # generate column names
        channel_name = self.channel_names[self.channel_selection[-1] - self.n_masks]
        self._generate_column_names(n_masks=self.n_masks, n_channels=1, channel_names=[channel_name])

        # define inference function
        f = func_partial(self.calculate_statistics, n_masks=self.n_masks)

        results = self.inference(
            self.dataloader,
            f,
            column_names=self.column_names,
        )
        if return_results:
            self._clear_cache()
            return results
        else:
            output_name = f"calculated_image_features_Channel_{channel_name}"
            path = os.path.join(self.run_path, f"{output_name}.csv")

            self._write_results_csv(results, path)
            self._write_results_sdata(results)

            # perform post processing cleanup
            if not self.deep_debug:
                self._post_processing_cleanup()
            self._clear_cache()
            return None



import multiprocessing
import os
import sys
import time
import timeit

import matplotlib.pyplot as plt
import numpy as np
import torch
import xarray
from alphabase.io import tempmmap
from cellpose import models
from skfmm import travel_time as skfmm_travel_time
from skimage.color import label2rgb
from skimage.filters import median
from skimage.morphology import binary_erosion, dilation, disk, erosion
from skimage.segmentation import watershed

from scportrait.pipeline._utils.segmentation import (
    contact_filter,
    global_otsu,
    numba_mask_centroid,
    remove_edge_labels,
    segment_global_threshold,
    segment_local_threshold,
)
from scportrait.pipeline.segmentation.segmentation import (
    Segmentation,
    ShardedSegmentation,
)
from scportrait.plotting._utils import _custom_cmap
from scportrait.processing.images._image_processing import downsample_img, percentile_normalization
from scportrait.processing.masks.mask_filtering import MatchNucleusCytosolIds, SizeFilter


class _BaseSegmentation(Segmentation):
    MASK_NAMES = ["nucleus", "cytosol"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._setup_channel_selection()
        self.nGPUs = None

    def _setup_channel_selection(self):
        self._setup_maximum_intensity_projection()
        self._define_channels_to_extract_for_segmentation()
        self._remap_maximum_intensity_projection_channels()

    def _setup_maximum_intensity_projection(self):
        # check if channels that should be maximum intensity projected and combined are defined in the config
        if "combine_cytosol_channels" in self.config.keys():
            self.combine_cytosol_channels = self.config["combine_cytosol_channels"]
            assert isinstance(
                self.combine_cytosol_channels, list
            ), "combine_cytosol_channels must be a list of integers specifying the indexes of the channels to combine."
            assert (
                len(self.combine_cytosol_channels) > 1
            ), "combine_cytosol_channels must contain at least two integers specifying the indexes of the channels to combine."
            self.maximum_project_cytosol = True
        else:
            self.combine_cytosol_channels = None
            self.maximum_project_cytosol = False

        if "combine_nucleus_channels" in self.config.keys():
            self.combine_nucleus_channels = self.config["combine_nucleus_channels"]
            assert isinstance(
                self.combine_nucleus_channels, list
            ), "combine_nucleus_channels must be a list of integers specifying the indexes of the channels to combine."
            assert (
                len(self.combine_nucleus_channels) > 1
            ), "combine_nucleus_channels must contain at least two integers specifying the indexes of the channels to combine."
            self.maximum_project_nucleus = True
        else:
            self.combine_nucleus_channels = None
            self.maximum_project_nucleus = False

    def _define_channels_to_extract_for_segmentation(self):
        self.segmentation_channels = []

        if "nucleus" in self.MASK_NAMES:
            if "segmentation_channel_nuclei" in self.config.keys():
                self.nucleus_segmentation_channel = self.config["segmentation_channel_nuclei"]
            elif "combine_nucleus_channels" in self.config.keys():
                self.nucleus_segmentation_channel = self.combine_nucleus_channels
            else:
                self.nucleus_segmentation_channel = self.DEFAULT_NUCLEI_CHANNEL_IDS

            self.segmentation_channels.extend(self.nucleus_segmentation_channel)

        if "cytosol" in self.MASK_NAMES:
            if "segmentation_channel_cytosol" in self.config.keys():
                self.cytosol_segmentation_channel = self.config["segmentation_channel_cytosol"]
            elif "combine_cytosol_channels" in self.config.keys():
                self.cytosol_segmentation_channel = self.combine_cytosol_channels
            else:
                self.cytosol_segmentation_channel = self.DEFAULT_CYTOSOL_CHANNEL_IDS

            self.segmentation_channels.extend(self.cytosol_segmentation_channel)

        # remove any duplicate entries and sort according to order
        self.segmentation_channels = list(set(self.segmentation_channels))

        # check validity of resulting list of segmentation channels
        assert len(self.segmentation_channels) > 0, "No segmentation channels specified in config file."
        assert (
            len(self.segmentation_channels) >= self.N_INPUT_CHANNELS
        ), f"Fewer segmentation channels {self.segmentation_channels} provided than expected by segmentation method {self.N_INPUT_CHANNELS}."

        if len(self.segmentation_channels) > self.N_INPUT_CHANNELS:
            assert (
                self.maximum_project_nucleus or self.maximum_project_cytosol
            ), "More input channels provided than accepted by the segmentation method and no maximum intensity projection performed on any of the input values."

    def _remap_maximum_intensity_projection_channels(self):
        """After selecting channels that are passed to the segmentation update indexes of the channels for maximum intensity projection so that they reflect the provided image subset"""
        if self.maximum_project_nucleus:
            self.original_combine_nucleus_channels = self.combine_nucleus_channels
            self.combine_nucleus_channels = [self.segmentation_channels.index(x) for x in self.combine_nucleus_channels]
        if self.maximum_project_cytosol:
            self.original_combine_cytosol_channels = self.combine_cytosol_channels
            self.combine_cytosol_channels = [self.segmentation_channels.index(x) for x in self.combine_cytosol_channels]

    def _transform_input_image(self, input_image):
        start_transform = timeit.default_timer()

        if isinstance(input_image, xarray.DataArray):
            input_image = input_image.data.compute()

        values = []
        # check if any channels need to be transformed
        if "nucleus" in self.MASK_NAMES:
            if self.maximum_project_nucleus:
                self.log(
                    f"For nucleus segmentation using the maximum intensity projection of channels {self.original_combine_nucleus_channels}."
                )
                nucleus_channel = self._maximum_project_channels(input_image, self.combine_nucleus_channels)
                if len(nucleus_channel.shape) == 2:
                    nucleus_channel = nucleus_channel[np.newaxis, ...]
                if len(nucleus_channel.shape) == 4:
                    nucleus_channel = nucleus_channel.squeeze()
            else:
                nucleus_channel = input_image[self.DEFAULT_NUCLEI_CHANNEL_IDS]
                if len(nucleus_channel.shape) == 2:
                    nucleus_channel = nucleus_channel[np.newaxis, ...]
                if len(nucleus_channel.shape) == 4:
                    nucleus_channel = nucleus_channel.squeeze()
            values.append(nucleus_channel)

        if "cytosol" in self.MASK_NAMES:
            if self.maximum_project_cytosol:
                self.log(
                    f"For cytosol segmentation using the maximum intensity projection of channels {self.original_combine_cytosol_channels}."
                )
                cytosol_channel = self._maximum_project_channels(input_image, self.combine_cytosol_channels)
                if len(cytosol_channel.shape) == 4:
                    cytosol_channel = cytosol_channel.squeeze()
                if len(cytosol_channel.shape) == 2:
                    cytosol_channel = cytosol_channel[np.newaxis, ...]
            else:
                cytosol_channel = input_image[self.DEFAULT_CYTOSOL_CHANNEL_IDS]
                if len(cytosol_channel.shape) == 4:
                    cytosol_channel = cytosol_channel.squeeze()
                if len(cytosol_channel.shape) == 2:
                    cytosol_channel = cytosol_channel[np.newaxis, ...]

            values.append(cytosol_channel)

        input_image = np.vstack(values)

        assert (
            input_image.shape[0] == self.N_INPUT_CHANNELS
        ), f"Number of channels in input image {input_image.shape[0]} does not match the number of channels expected by segmentation method {self.N_INPUT_CHANNELS}."

        stop_transform = timeit.default_timer()
        self.transform_time = stop_transform - start_transform
        return input_image

    def return_empty_mask(self, input_image):
        _, x, y = input_image.shape
        self._save_segmentation_sdata(np.zeros((self.N_MASKS, x, y)), [])

    def _check_seg_dtype(self, mask: np.ndarray, mask_name: str) -> np.ndarray:
        if not isinstance(mask, self.DEFAULT_SEGMENTATION_DTYPE):
            Warning(
                f"{mask_name} segmentation map is not of the correct dtype. \n Forcefully converting {mask.dtype} to {self.DEFAULT_SEGMENTATION_DTYPE}. \n This could lead to unexpected behaviour."
            )

            return mask.astype(self.DEFAULT_SEGMENTATION_DTYPE)

        else:
            return mask

    #### Downsampling ####
    def _get_downsampling_parameters(self) -> None:
        self.N = self.config["downsampling_factor"]

        if "smoothing_kernel_size" in self.config.keys():
            self.smoothing_kernel_size = self.config["smoothing_kernel_size"]

            if self.smoothing_kernel_size > self.N:
                self.log(
                    "Warning: Smoothing Kernel size is larger than the downsampling factor. This can lead to issues during smoothing where segmentation masks are lost. Please ensure to double check your results."
                )

        else:
            self.log(
                "Smoothing Kernel size not explicitly defined. Will calculate a default value based on the downsampling factor."
            )
            self.smoothing_kernel_size = self.N

        return None

    def _calculate_padded_image_size(self, img: np.array) -> None:
        """prepare metrics for image downsampling. Calculates image padding required for downsampling and returns
        metrics for this as well as resulting downsampled image size.
        """

        self.input_image_size = img.shape

        # check if N fits perfectly into image shape if not calculate how much we need to pad
        _, x, y = self.input_image_size

        if x % self.N == 0:
            pad_x = (0, 0)
        else:
            pad_x = (0, self.N - x % self.N)

        if y % self.N == 0:
            pad_y = (0, 0)
        else:
            pad_y = (0, self.N - y % self.N)

        # calculate resulting image size for use when e.g. inititalizing empty arrays to save results to
        padded_image_size = (
            2,
            self.input_image_size[1] + pad_x[1],
            self.input_image_size[2] + pad_y[1],
        )

        self.expected_padded_image_size = padded_image_size
        self.pad_x = pad_x
        self.pad_y = pad_y

        # track original image size
        self.original_image_size = img.shape

        return None

    def _downsample_image(self, img: np.array, debug: bool = False) -> np.array:
        """
        Downsample image by a factor of N. Before downsampling this function will pad the image to ensure its compatible with the selected kernel size.

        Parameters
        ----------
        img
            image to be downsampled

        Returns
        -------
        downsampled image

        """

        self.log(
            f"Performing image padding to ensure that image is compatible with selected downsample kernel size of {self.N}."
        )

        # perform image padding to ensure that image is compatible with downsample kernel size
        img = np.pad(img, ((0, 0), self.pad_x, self.pad_y))
        self.padded_image_size = img.shape

        if debug:
            self.log(
                f"Original image had size {self.expected_padded_image_size}, padded image is {self.padded_image_size}"
            )

        # sanity check to make sure padding worked as we wanted
        if self.expected_padded_image_size != self.padded_image_size:
            Warning(
                f"Expected a padded image of size {self.expected_padded_image_size} but got {self.padded_image_size}. Padding did not work as expted"
            )
            sys.exit(
                f"Error. Image padding did not work as expected. Expected a padded image of size {self.expected_padded_image_size} but got {self.padded_image_size}."
            )

        self.log(f"Downsampling image by a factor of {self.N}x{self.N}")

        # actually perform downsampling
        img = downsample_img(img, N=self.N)

        self.downsampled_image_size = img.shape

        if debug:
            self.log(f"Downsampled image size {self.downsampled_image_size}")

        return img

    def _rescale_downsampled_mask(self, mask: np.array, mask_name: str) -> np.array:
        input_mask = mask.copy()

        # get number of objects in mask for sanity checking
        n_classes = len(np.unique(mask))

        # rescale segmentations masks to padded image size
        mask = mask.repeat(self.N, axis=0).repeat(self.N, axis=1)

        # perform erosion and dilation for smoothing
        mask = erosion(mask, footprint=disk(self.smoothing_kernel_size))
        mask = dilation(
            mask, footprint=disk(self.smoothing_kernel_size + 1)
        )  # dilate 1 more than eroded to ensure that we do not lose any pixels

        # sanity check to make sure that smoothing does not remove masks
        if len(np.unique(mask)) != n_classes:
            Warning(
                "Number of objects in segmentation mask changed after smoothing. This should not happen. Ensure that you have chosen adequate smoothing parameters."
            )

            self.log(
                f"Will recalculate upsampling of {mask_name} mask with lower smoothing value to prevent the number of segmented objects from changing. Please ensure to double check your results."
            )

            smoothing_kernel_size = self.smoothing_kernel_size

            while len(np.unique(mask)) != n_classes:
                smoothing_kernel_size = smoothing_kernel_size - 1

                if smoothing_kernel_size == 0:
                    # if we reach 0 then we do not perform any smoothing
                    # repeat rescaling of the original mask

                    mask = input_mask
                    mask = mask.repeat(self.N, axis=0).repeat(self.N, axis=1)
                    self.log(f"Did not perform smoothing of {mask_name} mask.")

                    break

                else:
                    mask = input_mask
                    mask = mask.repeat(self.N, axis=0).repeat(self.N, axis=1)

                    # perform erosion and dilation for smoothing
                    mask = erosion(mask, footprint=disk(smoothing_kernel_size))

                    mask = dilation(
                        mask, footprint=disk(smoothing_kernel_size + 1)
                    )  # dilate 1 more than eroded to ensure that we do not lose any pixels

            self.log(
                f"Recalculation of {mask_name} mask successful with smoothing kernel size of {smoothing_kernel_size}."
            )

        # remove padding from mask
        x_trim = self.padded_image_size[1] - self.original_image_size[1]
        y_trim = self.padded_image_size[2] - self.original_image_size[2]

        # sanity check to ensure that we are removing what we addded
        assert x_trim == self.pad_x[1]
        assert y_trim == self.pad_y[1]

        # actually perform trimming

        if len(mask.shape) == 3:
            if x_trim > 0:
                if y_trim > 0:
                    mask = mask[:, :-x_trim, :-y_trim]
                else:
                    mask = mask[:, :-x_trim, :]
            else:
                if y_trim > 0:
                    mask = mask[:, :, :-y_trim]
                else:
                    mask = mask
        elif len(mask.shape) == 2:
            if x_trim > 0:
                if y_trim > 0:
                    mask = mask[:-x_trim, :-y_trim]
                else:
                    mask = mask[:-x_trim, :]
            else:
                if y_trim > 0:
                    mask = mask[:, :-y_trim]
                else:
                    mask = mask

        # check that mask has the correct shape and matches to input image

        if len(mask.shape) == 2:
            assert (mask.shape[0] == self.original_image_size[1]) and (mask.shape[1] == self.original_image_size[2])
        elif len(mask.shape) == 3:
            assert (mask.shape[1] == self.original_image_size[1]) and (mask.shape[2] == self.original_image_size[2])

        return mask

    #### Image Processing #####

    def _maximum_project_channels(self, input_image: np.array, channel_ids: list[int]) -> np.array:
        """add multiple channels together to generate a new channel for segmentation using maximum intensity projection"""

        dtype = input_image.dtype
        new_channel = np.zeros_like(input_image[0], dtype=dtype)

        for channel_id in channel_ids:
            new_channel = np.maximum(new_channel, input_image[channel_id])

        return new_channel

    def _normalize_image(
        self,
        input_image: np.array,
        lower: float | list,
        upper: float | list,
        debug: bool = False,
    ) -> np.array:
        if isinstance(lower, float) and isinstance(upper, float):
            self.log("Normalizing each channel to the same range")
            norm_image = percentile_normalization(input_image, lower, upper)

        elif isinstance(lower, list) and isinstance(upper, list):
            norm_image = []

            for i in range(input_image.shape[0]):
                _lower = lower[i]
                _upper = upper[i]  # type: ignore

                norm_image.append(percentile_normalization(input_image[i], _lower, _upper))

            norm_image = np.array(norm_image)
        else:
            raise ValueError(
                "Lower and upper quantile normalization values must be either floats or dictionary of floats."
            )
        if debug:
            if len(norm_image.shape) == 2:
                fig, ax = plt.subplots(1, 1, figsize=(10, 10))
                ax.imshow(norm_image)
                ax.axis("off")
                ax.set_title("Normalized Image")
            else:
                # visualize output if debug is turned on
                n_channels = norm_image.shape[0]
                fig, axs = plt.subplots(1, n_channels, figsize=(10 * n_channels, 10))
                for i in range(n_channels):
                    axs[i].imshow(norm_image[i])
                    axs[i].axis("off")
                    axs[i].set_title(f"Normalized Channel {i}")

            fig.tight_layout()
            fig_path = os.path.join(self.directory, "normalized_input.png")
            fig.savefig(fig_path)

        return norm_image

    def _median_correct_image(self, input_image, median_filter_size: int, debug: bool = False):
        self.log("Performing median filtering on input image")

        for i in range(input_image.shape[0]):
            input_image[i] = median(input_image[i], disk(median_filter_size))

        if debug:
            if len(input_image.shape) == 2:
                fig, ax = plt.subplots(1, 1, figsize=(10, 10))
                ax.imshow(input_image)
                ax.axis("off")
                ax.set_title("Median Corrected Image")
            else:
                # visualize output if debug is turned on
                n_channels = input_image.shape[0]

                fig, axs = plt.subplots(1, n_channels, figsize=(10 * n_channels, 10))
                for i in range(n_channels):
                    axs[i].imshow(input_image[i])
                    axs[i].axis("off")
                    axs[i].set_title(f"Median Corrected Channel {i}")

            fig.tight_layout()
            fig_path = os.path.join(self.directory, "median_corrected_input.png")
            fig.savefig(fig_path)

        return input_image

    ##### Filtering Functions #####

    # 1. Size Filtering
    def _check_for_size_filtering(self, mask_types: list[str]) -> None:
        """
        Check if size filtering should be performed on the masks.
        If size filtering is turned on, the thresholds for filtering are loaded from the config file.
        """

        assert all(
            mask_type in self.MASK_NAMES for mask_type in mask_types
        ), f"mask_types must be a list of strings that are valid mask names {self.MASK_NAMES}."

        if "filter_masks_size" in self.config.keys():
            self.filter_size = self.config["filter_masks_size"]
        else:
            # default behaviour is this is turned off filtering can always be performed later and this preserves the whole segmentation mask
            self.filter_size = False
            for mask_type in mask_types:
                # save attributes for use later
                setattr(self, f"{mask_type}_thresholds", None)
                setattr(self, f"{mask_type}_confidence_interval", None)

        # load parameters for cellsize filtering
        if self.filter_size:
            for mask_type in mask_types:
                thresholds, confidence_interval = self._get_params_cellsize_filtering(type=mask_type)

                # save attributes for use later
                setattr(self, f"{mask_type}_thresholds", thresholds)
                setattr(self, f"{mask_type}_confidence_interval", confidence_interval)

    def _get_params_cellsize_filtering(self, type) -> tuple[tuple[float] | None, float | None]:
        absolute_filter_status = False

        if "min_size" in self.config[f"{type}_segmentation"].keys():
            min_size = self.config[f"{type}_segmentation"]["min_size"]
            absolute_filter_status = True
        else:
            min_size = None

        if "max_size" in self.config[f"{type}_segmentation"].keys():
            max_size = self.config[f"{type}_segmentation"]["max_size"]
            absolute_filter_status = True
        else:
            max_size = None

        if absolute_filter_status:
            thresholds = [min_size, max_size]
            return (thresholds, None)  # type: ignore
        else:
            thresholds = None

            # get confidence intervals to automatically calculate thresholds
            if "confidence_interval" in self.config[f"{type}_segmentation"].keys():
                confidence_interval = self.config[f"{type}_segmentation"]["confidence_interval"]
            else:
                # get default value
                self.log(f"No confidence interval specified for {type} mask filtering, using default value of 0.95")
                confidence_interval = 0.95

            return (thresholds, confidence_interval)

    def _perform_size_filtering(
        self,
        mask: np.array,
        thresholds: tuple[float] | None,
        confidence_interval: float,
        mask_name: str,
        log: bool = True,
        debug: bool = False,
        input_image: np.array = None,
    ) -> np.array:
        """
        Remove elements from mask based on a size filter.

        Parameters
        ----------
        mask
            mask to be filtered
        """
        start_time = time.time()

        if thresholds is not None:
            self.log(f"Performing filtering of {mask_name} with specified thresholds {thresholds} from config file.")
        else:
            self.log(
                f"Automatically calculating thresholds for filtering of {mask_name} based on a fitted normal distribution with a confidence interval of {confidence_interval * 100}%."
            )

        filter = SizeFilter(
            label=mask_name,
            log=log,
            plot_qc=self.debug,
            directory=self.directory,
            confidence_interval=confidence_interval,
            filter_threshold=thresholds,
        )

        filtered_mask = filter.filter(mask)
        self.log(
            f"Removed {len(filter.ids_to_remove)} nuclei as they fell outside of the threshold range {filter.filter_threshold}."
        )

        if self.debug:
            # get visualization of the filtering results (2 = filtered, 1 = keep, 0 = background)
            mask = filter.visualize_filtering_results(plot_fig=False, return_maps=True)

            if not self.is_shard:
                if self.save_filter_results:
                    if len(mask.shape) == 2:
                        # add results to sdata object
                        self.filehandler._write_segmentation_sdata(
                            mask,
                            segmentation_label=f"debugging_seg_size_filter_results_{mask_name}",
                            classes=filter.ids,
                        )
                    else:
                        self.filehandler._write_segmentation_sdata(
                            mask[0],
                            segmentation_label=f"debugging_seg_size_filter_results_{mask_name}",
                            classes=filter.ids,
                        )
                    # then this does not need to be plotted as it can be visualized from there
                    plot_results = False
                else:
                    plot_results = True
            else:
                plot_results = True

            if plot_results:
                # get input image for visualization
                if input_image is not None:
                    if len(input_image.shape) == 2:
                        image_map = input_image
                    else:
                        if mask_name == "nucleus":
                            image_map = input_image[0]
                        elif mask_name == "cytosol":
                            image_map = input_image[1]

                    cmap, norm = _custom_cmap()

                    fig, axs = plt.subplots(1, 1, figsize=(10, 10))
                    axs.imshow(image_map, cmap="gray")
                    axs.imshow(mask, cmap=cmap, norm=norm)
                    axs.axis("off")
                    axs.set_title(f"Visualization of classes removed during {mask_name} size filtering")
                    fig_path = os.path.join(self.directory, f"Results_{mask_name}_size_filtering.png")
                    fig.savefig(fig_path)

                    self._clear_cache(vars_to_delete=[fig, input_image, cmap, norm])

            # clearup memory
            self._clear_cache(vars_to_delete=[mask, plot_results])

        end_time = time.time()

        self.log(f"Total time to perform {mask_name} size filtering: {end_time - start_time} seconds")

        return filtered_mask

    # 2. matching masks
    def _check_for_mask_matching_filtering(self) -> None:
        """Check to see if the masks should be filtered for matching nuclei/cytosols within the segmentation run."""

        DEFAULT_FILTERING_THRESHOLD_MASK_MATCHING = 0.95
        # check to see if the cells should be filtered for matching nuclei/cytosols within the segmentation run
        if "match_masks" in self.config.keys():
            self.filter_match_masks = self.config["match_masks"]
            if "filtering_threshold_mask_matching" in self.config.keys():
                self.mask_matching_filtering_threshold = self.config["filtering_threshold_mask_matching"]
            else:
                self.mask_matching_filtering_threshold = (
                    DEFAULT_FILTERING_THRESHOLD_MASK_MATCHING  # set default parameter
                )

        else:
            # default behaviour that this filtering should be performed, otherwise another additional step is required before extraction
            self.filter_match_masks = True
            self.mask_matching_filtering_threshold = DEFAULT_FILTERING_THRESHOLD_MASK_MATCHING

        # sanity check provided values
        assert isinstance(self.filter_match_masks, bool), "`match_masks` must be a boolean value."
        if self.filter_match_masks:
            assert isinstance(
                self.mask_matching_filtering_threshold, float
            ), "`filtering_threshold_mask_matching` for mask matching must be a float."

    def _perform_mask_matching_filtering(
        self,
        nucleus_mask: np.array,
        cytosol_mask: np.array,
        filtering_threshold: float,
        debug: bool = False,
        input_image: np.array = None,
    ) -> tuple[np.array, np.array]:
        """
        Match the nuclei and cytosol masks to ensure that the same cells are present in both masks.

        Parameters
        ----------
        nucleus_mask
            nucleus mask to be matched
        cytosol_mask
            cytosol mask to be matched
        """
        start_time = time.time()
        self.log("Performing filtering to match Cytosol and Nucleus IDs.")

        # perform filtering to remove cytosols which do not have a corresponding nucleus
        filter = MatchNucleusCytosolIds(filtering_threshold=filtering_threshold)
        masks_nucleus, masks_cytosol = filter.filter(nucleus_mask=nucleus_mask, cytosol_mask=cytosol_mask)

        self.log(
            f"Removed {len(filter.nuclei_discard_list)} nuclei and {len(filter.cytosol_discard_list)} cytosols due to filtering."
        )
        self.log(f"After filtering, {len(filter.nucleus_lookup_dict)} matching nuclei and cytosol masks remain.")

        if debug:
            mask_nuc, mask_cyto = filter.visualize_filtering_results(plot_fig=False, return_maps=True)

            # check if image should be added to sdata or plotted and saved
            if not self.is_shard:
                if self.save_filter_results:
                    # add filtering results to sdata object
                    self.filehandler._write_segmentation_sdata(
                        mask_nuc[0], segmentation_label="debugging_seg_match_mask_results_nucleus", classes=None
                    )
                    self.filehandler._write_segmentation_sdata(
                        mask_cyto[0], segmentation_label="debugging_seg_match_mask_result_cytosol", classes=None
                    )

                    # then no plotting needs to be performed as the results can be viewed in the sdata object
                    plot_results = False
                else:
                    plot_results = True
            else:
                plot_results = True

            if plot_results:
                if input_image is not None:
                    # convert input image from uint16 to uint8
                    input_image = (input_image / 256).astype(np.uint8)

                    cmap, norm = _custom_cmap()

                    fig, axs = plt.subplots(1, 2, figsize=(20, 10))

                    axs[0].imshow(input_image[0], cmap="gray")
                    axs[0].imshow(mask_nuc[0], cmap=cmap, norm=norm)
                    axs[0].imshow(mask_cyto[0], cmap=cmap, norm=norm)
                    axs[0].axis("off")
                    axs[0].set_title("results overlayed nucleus channel")

                    axs[1].imshow(input_image[1], cmap="gray")
                    axs[1].imshow(mask_nuc[0], cmap=cmap, norm=norm)
                    axs[1].imshow(mask_cyto[0], cmap=cmap, norm=norm)
                    axs[1].axis("off")
                    axs[1].set_title("results overlayed cytosol channel")

                    fig.tight_layout()
                    fig_path = os.path.join(self.directory, "Results_mask_matching.png")
                    fig.savefig(fig_path)

                    self._clear_cache(vars_to_delete=[fig, input_image, cmap, norm])

            # clearup memory
            self._clear_cache(vars_to_delete=[mask_nuc, mask_cyto, plot_results])

        self.log(
            f"Total time to perform nucleus and cytosol mask matching filtering: {time.time() - start_time:.2f} seconds"
        )

        return masks_nucleus, masks_cytosol


###### CLASSICAL SEGMENTATION METHODS #####


class _ClassicalSegmentation(_BaseSegmentation):
    def __init__(self, *args, **kwargs):
        self.maps = None  # just a typehint hack. Remove this and fix it properly.
        super().__init__(*args, **kwargs)

    def _visualize_final_masks(self):
        if self.segment_nuclei:
            image = label2rgb(
                self.maps["nucleus_segmentation"],
                self.maps["input_image"][0],
                bg_label=0,
                alpha=0.6,
            )

            fig = plt.figure(frameon=False, figsize=(10, 10))
            plt.imshow(image)
            plt.scatter(self.nucleus_centers[:, 1], self.nucleus_centers[:, 0], s=2, color="red")
            plt.axis("off")
            fig.tight_layout()
            fig.savefig(os.path.join(self.directory, "finalized_nucleus_mask.png"))

        if self.segment_cytosol:
            image = label2rgb(
                self.maps["cytosol_segmentation"],
                self.maps["input_image"][1],
                bg_label=0,
                alpha=0.6,
            )

            fig = plt.figure(frameon=False, figsize=(10, 10))
            plt.imshow(image)
            plt.scatter(self.nucleus_centers[:, 1], self.nucleus_centers[:, 0], s=2, color="red")
            plt.axis("off")
            fig.tight_layout()
            fig.savefig(os.path.join(self.directory, "cytosol_mask.png"))

    def _get_processing_parameters(self):
        # check if normalization is required
        if "lower_quantile_normalization" in self.config.keys():
            self.normalization_input_image = True
        elif "upper_quantile_normalization" in self.config.keys():
            self.normalization_input_image = True
        else:
            self.normalization_input_image = False
            self.log("No normalization will be performed")

        if self.normalization_input_image:
            self.log("Normalization of the input image will be performed.")

            # get lower quantile normalization range
            self.lower_quantile_normalization_input_image = 0
            self.upper_quantile_normalization_input_image = 1

            if isinstance(self.config["lower_quantile_normalization"], float):
                self.lower_quantile_normalization_input_image = self.config["lower_quantile_normalization"]

            elif isinstance(self.config["lower_quantile_normalization"], list):
                if len(self.config["lower_quantile_normalization"]) != self.input_image.shape[0]:
                    raise ValueError(
                        "Number of specified normalization ranges for lower quantile normalization does not match the number of input channels."
                    )
                self.lower_quantile_normalization_input_image = dict(
                    zip(
                        range(self.input_image.shape[0]),
                        self.config["lower_quantile_normalization"],
                        strict=False,
                    )
                )
            else:
                param = self.config["lower_quantile_normalization"]
                raise ValueError(
                    f"lower_quantile_normalization must be either a float or a list of floats. Instead recieved {param}"
                )

            # get upper quantile normalization range
            if isinstance(self.config["upper_quantile_normalization"], float):
                self.upper_quantile_normalization_input_image = self.config["upper_quantile_normalization"]

            elif isinstance(self.config["upper_quantile_normalization"], list):
                if len(self.config["upper_quantile_normalization"]) != self.input_image.shape[0]:
                    raise ValueError(
                        "Number of specified normalization ranges for upper quantile normalization does not match the number of input channels."
                    )
                self.upper_quantile_normalization_input_image = dict(
                    zip(
                        range(self.input_image.shape[0]),
                        self.config["upper_quantile_normalization"],
                        strict=False,
                    )
                )
            else:
                param = self.config["upper_quantile_normalization"]
                raise ValueError(
                    f"upper_quantile_normalization must be either a float or a list of floats. Instead recieved {param}"
                )

            # check that the normalization ranges are of the same type otherwise this will result in issues
            assert type(self.lower_quantile_normalization_input_image) == type(  # noqa: E721
                self.upper_quantile_normalization_input_image
            )  # these need to be the same types! So we need to circumvent the ruff linting rules here

        # check if median filtering is required
        if "median_filter_size" in self.config.keys():
            self.median_filtering_input_image = True
            self.median_filter_size_input_image = self.config["median_filter_size"]
        else:
            self.median_filtering_input_image = False

        # setup nucleus segmentation
        if "nucleus_segmentation" in self.config.keys():
            self.segment_nuclei = True

            # check for additional normalization before segmentation
            # set default values to 0 and 1 so no normalization occurs
            self.normalization_nucleus_lower_quantile = 0
            self.normalization_cytosol_upper_quantile = 1
            self.normalization_nucleus_segmentation = False

            if "lower_quantile_normalization" in self.config["nucleus_segmentation"]:
                self.normalization_nucleus_segmentation = True
                self.normalization_nucleus_lower_quantile = self.config["nucleus_segmentation"][
                    "lower_quantile_normalization"
                ]
            if "upper_quantile_normalization" in self.config["nucleus_segmentation"]:
                self.normalization_nucleus_segmentation = True
                self.normalization_nucleus_upper_quantile = self.config["nucleus_segmentation"][
                    "upper_quantile_normalization"
                ]

            # check if nuclei should be filtered based on size
            self._check_for_size_filtering(mask_types=["nucleus"])

            # check if nuclei should be filtered based on contact
            if "contact_filter" in self.config["nucleus_segmentation"]:
                self.contact_filter_nuclei = True
                self.contact_filter_nuclei_threshold = self.config["nucleus_segmentation"]["contact_filter"]
            else:
                self.contact_filter_nuclei = False

        if "cytosol_segmentation" in self.config.keys():
            self.segment_cytosol = True
            self.config_cytosol_segmentation = self.config["cytosol_segmentation"]
            # check for additional normalization before segmentation
            # set default values to 0 and 1 so no normalization occurs
            self.normalization_cytosol_lower_quantile = 0
            self.normalization_cytosol_upper_quantile = 1
            self.normalization_cytosol_segmentation = False

            if "lower_quantile_normalization" in self.config["cytosol_segmentation"]:
                self.normalization_cytosol_segmentation = True
                self.normalization_cytosol_lower_quantile = self.config["cytosol_segmentation"][
                    "lower_quantile_normalization"
                ]
            if "upper_quantile_normalization" in self.config_cytosol_segmentation.keys():
                self.normalization_cytosol_segmentation = True
                self.normalization_cytosol_upper_quantile = self.config["cytosol_segmentation"][
                    "upper_quantile_normalization"
                ]

            # check if cytosol should be filtered based on size
            self._check_for_size_filtering(mask_types=["cytosol"])

            # check if cytosol should be filtered based on contact
            if "contact_filter" in self.config["cytosol_segmentation"]:
                self.contact_filter_cytosol = True
                self.contact_filter_cytosol_threshold = self.config["cytosol_segmentation"]["contact_filter"]
            else:
                self.contact_filter_cytosol = False

            # check if cytosol should be filtered based on matching to nuclei
            self._check_for_mask_matching_filtering()

    def _nucleus_segmentation(self, input_image, debug: bool = False):
        if self.normalization_nucleus_segmentation:
            lower = self.normalization_nucleus_lower_quantile
            upper = self.normalization_nucleus_upper_quantile

            self.log(f"Percentile normalization of nucleus input image to range {lower}, {upper}")
            input_image = self._normalize_image(input_image, lower, upper, debug=debug)

        # perform thresholding to generate  a mask of the nuclei

        # Use manual threshold if defined in ["wga_segmentation"]["threshold"]
        # If not, use global otsu
        if "threshold" in self.config["nucleus_segmentation"] and "median_block" in self.config["nucleus_segmentation"]:
            threshold = self.config["nucleus_segmentation"]["threshold"]
            self.log(f"Using local thresholding with a threshold of {threshold} to calculate nucleus mask.")
            self.maps["nucleus_segmentation"] = segment_local_threshold(
                input_image,
                dilation=self.config["nucleus_segmentation"]["dilation"],
                thr=self.config["nucleus_segmentation"]["threshold"],
                median_block=self.config["nucleus_segmentation"]["median_block"],
                min_distance=self.config["nucleus_segmentation"]["min_distance"],
                peak_footprint=self.config["nucleus_segmentation"]["peak_footprint"],
                speckle_kernel=self.config["nucleus_segmentation"]["speckle_kernel"],
                median_step=self.config["nucleus_segmentation"]["median_step"],
                debug=self.debug,
            )
        else:
            self.log("Using global otsu to calculate threshold for nucleus mask generation.")
            self.maps["nucleus_segmentation"] = segment_global_threshold(
                input_image,
                dilation=self.config["nucleus_segmentation"]["dilation"],
                min_distance=self.config["nucleus_segmentation"]["min_distance"],
                peak_footprint=self.config["nucleus_segmentation"]["peak_footprint"],
                speckle_kernel=self.config["nucleus_segmentation"]["speckle_kernel"],
                debug=self.debug,
            )

        # save the thresholding approach as a mask
        self.maps["nucleus_mask"] = np.clip(self.maps["nucleus_segmentation"], 0, 1)

        # calculate nucleus centroids
        # important to do this before any filtering! Otherwise we wont know where to start looking for cytosols
        centers, _, _ = numba_mask_centroid(self.maps["nucleus_segmentation"])
        self.nucleus_centers = centers

        # ensure all edge labels are removed
        self.maps["nucleus_segmentation"] = remove_edge_labels(self.maps["nucleus_segmentation"])

        if self.filter_size:
            self.maps["nucleus_segmentation"] = self._perform_size_filtering(
                self.maps["nucleus_segmentation"],
                self.nucleus_thresholds,  # type: ignore
                self.nucleus_confidence_interval,  # type: ignore
                "nucleus",
                debug=self.debug,
                input_image=input_image if self.debug else None,
            )

        if self.contact_filter_nuclei:
            if self.debug:
                n_classes = len(set(np.unique(self.maps["nucleus_segmentation"])) - {0})

            self.maps["nucleus_segmentation"] = contact_filter(
                self.maps["nucleus_segmentation"],
                threshold=self.contact_filter_nuclei_threshold,
                reindex=False,
            )

            if self.debug:
                n_classes_post = len(set(np.unique(self.maps["nucleus_segmentation"])) - {0})
                self.log(f"Filtered out {n_classes - n_classes_post} nuclei due to contact filtering.")

    def _cytosol_segmentation(self, input_image, debug: bool = False):
        if not self.segment_nuclei:
            raise ValueError("Nucleus segmentation must be performed to be able to perform a cytosol segmentation.")

        if self.normalization_cytosol_segmentation:
            lower = self.normalization_cytosol_lower_quantile
            upper = self.normalization_cytosol_upper_quantile

            input_image = self._normalize_image(input_image, lower, upper, debug=debug)

        # perform thresholding to generate a mask of the cytosol

        # Use manual threshold if defined in ["wga_segmentation"]["threshold"]
        # If not, use global otsu
        if "threshold" in self.config["cytosol_segmentation"]:
            cytosol_mask = input_image < self.config["cytosol_segmentation"]["threshold"]
        else:
            self.log("No treshold for cytosol segmentation defined, global otsu will be used.")
            cytosol_mask = input_image < global_otsu(input_image)

        self._clear_cache(vars_to_delete=[input_image])

        # remove the nucleus mask from the cytosol mask
        cytosol_mask = cytosol_mask.astype(float)
        cytosol_mask -= self.maps["nucleus_mask"]
        cytosol_mask = np.clip(cytosol_mask, 0, 1)

        # Apply dilation and erosion
        cytosol_mask = dilation(cytosol_mask, footprint=disk(self.config["cytosol_segmentation"]["erosion"]))
        cytosol_mask = binary_erosion(
            cytosol_mask,
            footprint=disk(self.config["cytosol_segmentation"]["dilation"]),
        )

        self.maps["cytosol_mask"] = cytosol_mask

        if self.debug:
            fig = plt.figure(figsize=(10, 10))
            plt.imshow(self.maps["cytosol_mask"], cmap="magma")
            plt.axis("off")
            plt.title("Cytosol Mask")
            fig_path = os.path.join(self.directory, "cytosol_mask.png")
            fig.savefig(fig_path)

        self._clear_cache(vars_to_delete=[cytosol_mask])

        # calculate potential map for cytosol segmentation
        potential_map = (
            self.maps["median_corrected"][1] - np.quantile(self.maps["median_corrected"][1], 0.02)
        ) / np.quantile(self.maps["median_corrected"][1], 0.98)
        potential_map = np.clip(potential_map, 0, 1)

        # subtract nucleus mask from potential map
        potential_map = np.clip(potential_map - self.maps["nucleus_mask"], 0, 1)
        potential_map = 1 - potential_map

        # enhance potential map to generate speedmap
        min_clip = self.config["cytosol_segmentation"]["min_clip"]
        max_clip = self.config["cytosol_segmentation"]["max_clip"]
        potential_map = (np.clip(potential_map, min_clip, max_clip) - min_clip) / (max_clip - min_clip)
        potential_map = (potential_map * 0.9 + 0.1).astype(float)

        self.maps["potential_map"] = potential_map

        if self.debug:
            fig = plt.figure(figsize=(10, 10))
            plt.imshow(self.maps["potential_map"], cmap="magma")
            plt.axis("off")
            plt.title("Potential Map")
            fig_path = os.path.join(self.directory, "potential_map.png")
            fig.savefig(fig_path)

        self.log("Cytosol Potential Mask generated.")
        self._clear_cache(vars_to_delete=[potential_map])

        # perform fast marching and watershed segmentation
        fmm_marker = np.ones_like(self.maps["median_corrected"][0])
        px_center = np.round(self.nucleus_centers).astype(np.uint64)

        for center in px_center:
            fmm_marker[center[0], center[1]] = 0

        fmm_marker = np.ma.MaskedArray(fmm_marker, self.maps["cytosol_mask"])
        travel_time = skfmm_travel_time(fmm_marker, self.maps["potential_map"])

        if not isinstance(travel_time, np.ma.core.MaskedArray):
            raise TypeError(
                "Travel_time for WGA based segmentation returned no MaskedArray. This is most likely due to missing WGA background determination."
            )

        self.maps["travel_time"] = travel_time.filled(fill_value=np.max(travel_time))

        marker = np.zeros_like(self.maps["median_corrected"][1])

        for center in px_center:
            marker[center[0], center[1]] = self.maps["nucleus_segmentation"][center[0], center[1]]

        cytosol_labels = watershed(
            self.maps["travel_time"],
            marker.astype(np.int64),
            mask=(self.maps["cytosol_mask"] == 0).astype(np.int64),
        )

        cytosol_segmentation = np.where(self.maps["cytosol_mask"] > 0.5, 0, cytosol_labels)

        # ensure all edge labels are removed
        cytosol_segmentation = remove_edge_labels(cytosol_segmentation)
        self.maps["cytosol_segmentation"] = cytosol_segmentation

        if self.filter_size:
            self.maps["cytosol_segmentation"] = self._perform_size_filtering(
                self.maps["cytosol_segmentation"],
                self.cytosol_thresholds,  # type: ignore
                self.cytosol_confidence_interval,  # type: ignore
                "cytosol",
                debug=self.debug,
                input_image=input_image if self.debug else None,
            )

        if self.contact_filter_cytosol:
            if self.debug:
                n_classes = len(set(np.unique(self.maps["cytosol_segmentation"])) - {0})

            self.maps["cytosol_segmentation"] = contact_filter(
                self.maps["cytosol_segmentation"],
                threshold=self.contact_filter_cytosol_threshold,
                reindex=False,
            )

            if self.debug:
                n_classes_post = len(set(np.unique(self.maps["cytosol_segmentation"])) - {0})
                self.log(f"Filtered out {n_classes - n_classes_post} cytosols due to contact filtering.")

        unique_cytosol_ids = set(np.unique(self.maps["cytosol_segmentation"])) - {0}

        # remove any ids from nucleus mask that dont have a cytosol mask
        self.maps["nucleus_segmentation"][~np.isin(self.maps["nucleus_segmentation"], list(unique_cytosol_ids))] = 0


class WGASegmentation(_ClassicalSegmentation):
    N_MASKS = 2
    N_INPUT_CHANNELS = 2
    MASK_NAMES = ["nucleus", "cytosol"]
    DEFAULT_NUCLEI_CHANNEL_IDS = [0]
    DEFAULT_CYTOSOL_CHANNEL_IDS = [1]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _finalize_segmentation_results(self):
        segmentation = np.stack([self.maps["nucleus_segmentation"], self.maps["cytosol_segmentation"]]).astype(
            self.DEFAULT_SEGMENTATION_DTYPE
        )

        return segmentation

    def _execute_segmentation(self, input_image):
        total_time_start = timeit.default_timer()
        self._get_processing_parameters()

        # intialize maps for storing intermediate results
        self.maps = {}

        input_image = self._transform_input_image(input_image)

        self.maps["input_image"] = input_image

        # normalize input
        if self.normalization_input_image:
            self.maps["normalized"] = self._normalize_image(
                input_image,
                self.lower_quantile_normalization_input_image,
                self.upper_quantile_normalization_input_image,
                debug=self.debug,
            )
            # update input image to normalized image
            input_image = self.maps["normalized"]

        if self.median_filtering_input_image:
            self.maps["median_corrected"] = self._median_correct_image(
                input_image, self.median_filter_size_input_image, debug=self.debug
            )
            # update input image to median corrected image
            input_image = self.maps["median_corrected"]

        start_segmentation = timeit.default_timer()
        if self.segment_nuclei:
            image = input_image[0]
            self._nucleus_segmentation(image, debug=self.debug)

        if self.segment_cytosol:
            image = self.maps["input_image"][1]
            self._cytosol_segmentation(image, debug=self.debug)
        stop_segmentation = timeit.default_timer()
        self.segmentation_time = stop_segmentation - start_segmentation

        if self.debug:
            self._visualize_final_masks()

        all_classes = list(set(np.unique(self.maps["nucleus_segmentation"])) - {0})
        segmentation = self._finalize_segmentation_results()  # type: ignore

        self._save_segmentation_sdata(segmentation, all_classes, masks=self.MASK_NAMES)
        self.total_time = timeit.default_timer() - total_time_start


class ShardedWGASegmentation(ShardedSegmentation):
    method = WGASegmentation

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class DAPISegmentation(_ClassicalSegmentation):
    N_MASKS = 1
    N_INPUT_CHANNELS = 1
    MASK_NAMES = ["nucleus"]
    DEFAULT_NUCLEI_CHANNEL_IDS = [0]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _finalize_segmentation_results(self):
        segmentation = np.stack([self.maps["nucleus_segmentation"]]).astype(self.DEFAULT_SEGMENTATION_DTYPE)

        return segmentation

    def _execute_segmentation(self, input_image):
        total_time_start = timeit.default_timer()
        self._get_processing_parameters()

        # intialize maps for storing intermediate results
        self.maps = {}

        input_image = self._transform_input_image(input_image)
        self.maps["input_image"] = input_image.copy()

        # normalize input
        if self.normalization_input_image:
            self.maps["normalized"] = self._normalize_image(
                input_image,
                self.lower_quantile_normalization_input_image,
                self.upper_quantile_normalization_input_image,
                debug=self.debug,
            )
            # update input image to normalized image
            input_image = self.maps["normalized"]

        if self.median_filtering_input_image:
            self.maps["median_corrected"] = self._median_correct_image(
                input_image, self.median_filter_size_input_image, debug=self.debug
            )
            # update input image to median corrected image
            input_image = self.maps["median_corrected"]

        start_segmentation = timeit.default_timer()
        if self.segment_nuclei:
            self._nucleus_segmentation(input_image[0], debug=self.debug)
        stop_segmentation = timeit.default_timer()
        self.segmentation_time = stop_segmentation - start_segmentation

        all_classes = list(set(np.unique(self.maps["nucleus_segmentation"])) - {0})
        segmentation = self._finalize_segmentation_results()

        self._save_segmentation_sdata(segmentation, all_classes, masks=self.MASK_NAMES)
        self.total_time = timeit.default_timer() - total_time_start


class ShardedDAPISegmentation(ShardedSegmentation):
    method = DAPISegmentation


##### CELLPOSE BASED SEGMENTATION METHODS #####


class _CellposeSegmentation(_BaseSegmentation):
    def _write_cellpose_seg_params_to_file(self, model_type: str, model_name: str) -> None:
        """
        Writes the cellpose segmentation parameters to a file for debugging/logging purposes
        """
        with open(os.path.join(self.directory, f"cellpose_params_{model_type}.txt"), "w") as f:
            f.write("Cellpose Parameters:\n")
            f.write(f"Model: {model_name}\n")
            f.write(f"Model Type: {model_type}\n")
            f.write(f"Diameter: {self.diameter}\n")
            f.write(f"Resample: {self.resample}\n")
            f.write(f"Flow Threshold: {self.flow_threshold}\n")
            f.write(f"Cellprob Threshold: {self.cellprob_threshold}\n")
            f.write(f"Normalize: {self.normalize}\n")
            f.write(f"Rescale: {self.rescale}\n")

    def _read_cellpose_model(self, modeltype: str, name: str, gpu: str, device) -> models.Cellpose:
        """
        Reads cellpose model based on the modeltype and name. Will load to GPU if available as specified in self._use_gpu

        Parameters
        ----------
        modeltype
            either "pretrained" or "custom" depending on the model to load
        name
            name of the model to load

        Returns
        -------
        cellpose model

        """
        if modeltype == "pretrained":
            model = models.Cellpose(model_type=name, gpu=gpu, device=device)
        elif modeltype == "custom":
            model = models.CellposeModel(pretrained_model=name, gpu=gpu, device=device)
        return model

    def _load_model(self, model_type: str, gpu: str, device) -> models.Cellpose:
        """
        Loads cellpose model

        Parameters
        ----------
        model_type
            either "cytosol" or "nucleus" depending on the model to load

        Returns
        -------
        tuple of expected diameter and the cellpose model
        """

        # load correct segmentation model for cytosol
        if "model" in self.config[f"{model_type}_segmentation"].keys():
            model_name = self.config[f"{model_type}_segmentation"]["model"]
            model = self._read_cellpose_model("pretrained", model_name, gpu=gpu, device=device)

        elif "model_path" in self.config[f"{model_type}_segmentation"].keys():
            model_name = self.config[f"{model_type}_segmentation"]["model_path"]
            model = self._read_cellpose_model("custom", model_name, gpu=gpu, device=device)

        # get model parameters from config if not defined use default values
        if "diameter" in self.config[f"{model_type}_segmentation"].keys():
            self.diameter = self.config[f"{model_type}_segmentation"]["diameter"]
        else:
            self.diameter = None

        if "resample" in self.config[f"{model_type}_segmentation"].keys():
            self.resample = self.config[f"{model_type}_segmentation"]["resample"]
        else:
            self.resample = True

        if "flow_threshold" in self.config[f"{model_type}_segmentation"].keys():
            self.flow_threshold = self.config[f"{model_type}_segmentation"]["flow_threshold"]
        else:
            self.flow_threshold = 0.4

        if "cellprob_threshold" in self.config[f"{model_type}_segmentation"].keys():
            self.cellprob_threshold = self.config[f"{model_type}_segmentation"]["cellprob_threshold"]
        else:
            self.cellprob_threshold = 0.0

        if "normalize" in self.config[f"{model_type}_segmentation"].keys():
            self.normalize = self.config[f"{model_type}_segmentation"]["normalize"]
        else:
            self.normalize = True

        if "rescale" in self.config[f"{model_type}_segmentation"].keys():
            self.rescale = self.config[f"{model_type}_segmentation"]["rescale"]
        else:
            self.rescale = None

        self.log(f"Segmenting {model_type} using the following model: {model_name}")

        self._write_cellpose_seg_params_to_file(model_type=model_type, model_name=model_name)

        return model

    def _check_input_image_dtype(self, input_image: np.ndarray):
        if input_image.dtype != self.DEFAULT_IMAGE_DTYPE:
            if isinstance(input_image.dtype, int):
                ValueError(
                    "Default image dtype is no longer int. Cellpose expects int inputs. Please contact developers."
                )
            else:
                ValueError("Image is not of type uint16, cellpose segmentation expects int input images.")

    def _check_gpu_status(self):
        """
        Checks and updates the GPU status.
        If a multi-GPU setup is used, the function checks the current process and returns the GPU id to use for the segmentation.
        If no GPUs are available, the function defaults to CPU.
        """

        # get GPU information if run with workers
        try:
            current = multiprocessing.current_process()
            cpu_name = current.name
            gpu_id_list = current.gpu_id_list
            cpu_id = int(cpu_name[cpu_name.find("-") + 1 :]) - 1

            if cpu_id >= len(gpu_id_list):
                cpu_id = cpu_id % current.n_processes

            # track gpu_id and update GPU status
            self.gpu_id = gpu_id_list[cpu_id]
            self.status = "multi_GPU"

        except (AttributeError, ValueError):
            # default to single GPU
            self.gpu_id = 0
            self.status = "potentially_single_GPU"

        # check if cuda GPU is available
        if torch.cuda.is_available():
            if self.status == "multi_GPU":
                self.use_GPU = f"cuda:{self.gpu_id}"
                self.device = torch.device(self.use_GPU)
                self.nGPUs = torch.cuda.device_count()
            else:
                self.use_GPU = True
                self.device = torch.device(
                    "cuda"
                )  # dont need to specify id, saying cuda will default to the one thats avaialable
                self.nGPUs = 1

        # check if MPS is available
        elif torch.backends.mps.is_available():
            self.use_GPU = True
            self.device = torch.device("mps")
            self.nGPUs = 1

        # default to CPU
        else:
            self.use_GPU = False
            self.device = torch.device("cpu")
            self.nGPUs = 0

        self.log(
            f"GPU Status for segmentation is {self.use_GPU} and will segment using the following device {self.device}."
        )


class DAPISegmentationCellpose(_CellposeSegmentation):
    N_MASKS = 1
    N_INPUT_CHANNELS = 1
    MASK_NAMES = ["nucleus"]
    DEFAULT_NUCLEI_CHANNEL_IDS = [0]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _setup_filtering(self):
        self._check_for_size_filtering(mask_types=self.MASK_NAMES)

    def _finalize_segmentation_results(self, nucleus_mask: np.ndarray) -> np.ndarray:
        # ensure correct dtype of the maps

        nucleus_mask = self._check_seg_dtype(mask=nucleus_mask, mask_name="nucleus")

        segmentation = np.stack([nucleus_mask])

        return segmentation

    def cellpose_segmentation(self, input_image: np.ndarray) -> np.ndarray:
        self._check_gpu_status()
        self._clear_cache()  # ensure we start with an empty cache

        ################################
        ### Perform Nucleus Segmentation
        ################################

        model = self._load_model(model_type="nucleus", gpu=self.use_GPU, device=self.device)

        if self.normalize is False:
            input_image = (input_image - np.min(input_image)) / (
                np.max(input_image) - np.min(input_image)
            )  # min max normalize to 0-1 range as cellpose expects this

        masks = model.eval(
            [input_image],
            rescale=self.rescale,
            normalize=self.normalize,
            diameter=self.diameter,
            flow_threshold=self.flow_threshold,
            cellprob_threshold=self.cellprob_threshold,
            channels=[1, 0],
        )[0]
        masks = np.array(masks)

        # ensure all edge classes are removed
        masks = remove_edge_labels(masks)

        # check if filtering is required
        self._setup_filtering()

        if self.filter_size:
            masks = self._perform_size_filtering(
                mask=masks,
                thresholds=self.nucleus_thresholds,  # type: ignore
                confidence_interval=self.nucleus_confidence_interval,  # type: ignore
                mask_name="nucleus",
                log=True,
                input_image=input_image if self.debug else None,
            )

        masks = masks.reshape(masks.shape[1:])
        return masks

    def _execute_segmentation(self, input_image):
        total_time_start = timeit.default_timer()

        # check that the correct level of input image is used
        input_image = self._transform_input_image(input_image)

        self._check_input_image_dtype(input_image)

        start_segmentation = timeit.default_timer()
        nucleus_mask = self.cellpose_segmentation(input_image)
        stop_segmentation = timeit.default_timer()
        self.segmentation_time = stop_segmentation - start_segmentation

        # finalize classes list
        all_classes = set(np.unique(nucleus_mask)) - {0}

        segmentation = self._finalize_segmentation_results(nucleus_mask=nucleus_mask)
        self._save_segmentation_sdata(segmentation, all_classes, masks=self.MASK_NAMES)
        self.total_time = timeit.default_timer() - total_time_start


class ShardedDAPISegmentationCellpose(ShardedSegmentation):
    method = DAPISegmentationCellpose


class CytosolSegmentationCellpose(_CellposeSegmentation):
    N_MASKS = 2
    N_INPUT_CHANNELS = 2
    MASK_NAMES = ["nucleus", "cytosol"]
    DEFAULT_NUCLEI_CHANNEL_IDS = [0]
    DEFAULT_CYTOSOL_CHANNEL_IDS = [1]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _finalize_segmentation_results(self, mask_nucleus: np.ndarray, mask_cytosol: np.ndarray) -> np.ndarray:
        # ensure correct dtype of maps

        mask_nucleus = self._check_seg_dtype(mask=mask_nucleus, mask_name="nucleus")
        mask_cytosol = self._check_seg_dtype(mask=mask_cytosol, mask_name="cytosol")

        segmentation = np.stack([mask_nucleus, mask_cytosol])
        return segmentation

    def _setup_filtering(self):
        self._check_for_size_filtering(mask_types=self.MASK_NAMES)
        self._check_for_mask_matching_filtering()

    def cellpose_segmentation(self, input_image: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
        self._check_gpu_status()
        self._clear_cache()  # ensure we start with an empty cache

        ################################
        ### Perform Nucleus Segmentation
        ################################

        model = self._load_model(model_type="nucleus", gpu=self.use_GPU, device=self.device)

        if self.normalize is False:
            input_image = (input_image - np.min(input_image)) / (
                np.max(input_image) - np.min(input_image)
            )  # min max normalize to 0-1 range as cellpose expects this

        masks_nucleus = model.eval(
            [input_image],
            rescale=self.rescale,
            normalize=self.normalize,
            diameter=self.diameter,
            flow_threshold=self.flow_threshold,
            cellprob_threshold=self.cellprob_threshold,
            channels=[1, 0],
        )[0]
        masks_nucleus = np.array(masks_nucleus)  # convert to array

        # manually delete model and perform gc to free up memory on GPU
        self._clear_cache(vars_to_delete=[model])

        # remove edge labels from masks
        masks_nucleus = remove_edge_labels(masks_nucleus)

        #################################
        #### Perform Cytosol Segmentation
        #################################

        model = self._load_model(model_type="cytosol", gpu=self.use_GPU, device=self.device)

        masks_cytosol = model.eval(
            [input_image],
            rescale=self.rescale,
            normalize=self.normalize,
            diameter=self.diameter,
            flow_threshold=self.flow_threshold,
            cellprob_threshold=self.cellprob_threshold,
            channels=[2, 1],
        )[0]
        masks_cytosol = np.array(masks_cytosol)  # convert to array

        # manually delete model and perform gc to free up memory on GPU
        self._clear_cache(vars_to_delete=[model])

        # remove edge labels from masks
        masks_cytosol = remove_edge_labels(masks_cytosol)

        ######################
        ### Perform Filtering to remove too small/too large masks if applicable
        ######################

        # check if filtering is required
        self._setup_filtering()

        if self.filter_size:
            masks_nucleus = self._perform_size_filtering(
                mask=masks_nucleus,
                thresholds=self.nucleus_thresholds,  # type: ignore
                confidence_interval=self.nucleus_confidence_interval,  # type: ignore
                mask_name="nucleus",
                log=True,
                debug=self.debug,
                input_image=input_image if self.debug else None,
            )

            masks_cytosol = self._perform_size_filtering(
                mask=masks_cytosol,
                thresholds=self.nucleus_thresholds,  # type: ignore
                confidence_interval=self.nucleus_confidence_interval,  # type: ignore
                mask_name="cytosol",
                log=True,
                debug=self.debug,
                input_image=input_image if self.debug else None,
            )

        ######################
        ### Perform Filtering match cytosol and nucleus IDs if applicable
        ######################
        self._setup_filtering()

        if self.filter_match_masks:
            masks_nucleus, masks_cytosol = self._perform_mask_matching_filtering(
                nucleus_mask=masks_nucleus,
                cytosol_mask=masks_cytosol,
                input_image=input_image if self.debug else None,
                filtering_threshold=self.mask_matching_filtering_threshold,
                debug=self.debug,
            )
        else:
            self.log(
                "No filtering performed. Cytosol and Nucleus IDs in the two masks do not match. Before proceeding with extraction an additional filtering step needs to be performed"
            )

        ######################
        ### Cleanup Generated Segmentation masks
        ######################

        masks_nucleus = masks_nucleus.reshape(masks_nucleus.shape[1:])
        masks_cytosol = masks_cytosol.reshape(masks_cytosol.shape[1:])

        return (masks_nucleus, masks_cytosol)

    def _execute_segmentation(self, input_image):
        total_time_start = timeit.default_timer()

        # ensure the correct level is selected for the input image
        input_image = self._transform_input_image(input_image)

        # check image dtype since cellpose expects int input images
        self._check_input_image_dtype(input_image)

        start_segmentation = timeit.default_timer()
        masks_nucleus, masks_cytosol = self.cellpose_segmentation(input_image)
        stop_segmentation = timeit.default_timer()
        self.segmentation_time = stop_segmentation - start_segmentation

        # finalize segmentation classes ensuring that background is removed
        all_classes = set(np.unique(masks_nucleus)) - {0}

        segmentation = self._finalize_segmentation_results(mask_nucleus=masks_nucleus, mask_cytosol=masks_cytosol)
        self._save_segmentation_sdata(segmentation, all_classes, masks=self.MASK_NAMES)

        # clean up memory
        self._clear_cache(vars_to_delete=[segmentation, all_classes])
        self.total_time = timeit.default_timer() - total_time_start


class ShardedCytosolSegmentationCellpose(ShardedSegmentation):
    method = CytosolSegmentationCellpose


class CytosolSegmentationDownsamplingCellpose(CytosolSegmentationCellpose):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _finalize_segmentation_results(self, mask_nucleus: np.ndarray, mask_cytosol: np.ndarray) -> np.ndarray:
        mask_nucleus = self._rescale_downsampled_mask(mask_nucleus, "nucleus_segmentation")
        mask_cytosol = self._rescale_downsampled_mask(mask_cytosol, "cytosol_segmentation")

        mask_nucleus = self._check_seg_dtype(mask=mask_nucleus, mask_name="nucleus")
        mask_cytosol = self._check_seg_dtype(mask=mask_cytosol, mask_name="cytosol")

        # combine masks into one stack
        segmentation = np.stack([mask_nucleus, mask_cytosol])

        return segmentation

    def _execute_segmentation(self, input_image):
        total_time_start = timeit.default_timer()

        # ensure the correct level is selected for the input image
        input_image = self._transform_input_image(input_image)

        # check image dtype since cellpose expects int input images
        self._check_input_image_dtype(input_image)

        # only get the first two channels to save memory consumption
        input_image = input_image[:2, :, :]

        # setup downsampling
        self._get_downsampling_parameters()
        self._calculate_padded_image_size(input_image)

        # downsample the image
        input_image = self._downsample_image(input_image)

        start_segmentation = timeit.default_timer()
        mask_nucleus, mask_cytosol = self.cellpose_segmentation(input_image)
        stop_segmentation = timeit.default_timer()
        self.segmentation_time = stop_segmentation - start_segmentation

        # finalize classes list
        all_classes = set(np.unique(mask_nucleus)) - {0}

        segmentation = self._finalize_segmentation_results(mask_nucleus=mask_nucleus, mask_cytosol=mask_cytosol)

        self._save_segmentation_sdata(segmentation, all_classes, masks=self.MASK_NAMES)
        self._clear_cache(vars_to_delete=[segmentation, all_classes])
        self.total_time = timeit.default_timer() - total_time_start


class ShardedCytosolSegmentationDownsamplingCellpose(ShardedSegmentation):
    method = CytosolSegmentationDownsamplingCellpose


class CytosolOnlySegmentationCellpose(_CellposeSegmentation):
    N_MASKS = 1
    N_INPUT_CHANNELS = 2
    MASK_NAMES = ["cytosol"]
    DEFAULT_CYTOSOL_CHANNEL_IDS = [0, 1]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _setup_filtering(self):
        self._check_for_size_filtering(mask_types=self.MASK_NAMES)

    def _finalize_segmentation_results(self, cytosol_mask: np.ndarray) -> np.ndarray:
        # ensure correct dtype of maps
        cytosol_mask = self._check_seg_dtype(mask=cytosol_mask, mask_name="cytosol")

        segmentation = np.stack([cytosol_mask])
        return segmentation

    def cellpose_segmentation(self, input_image: np.ndarray) -> np.ndarray:
        self._setup_processing()
        self._clear_cache()

        #####
        ### Perform Cytosol Segmentation
        #####

        model = self._load_model(model_type="cytosol", gpu=self.use_GPU, device=self.device)

        if self.normalize is False:
            input_image = (input_image - np.min(input_image)) / (
                np.max(input_image) - np.min(input_image)
            )  # min max normalize to 0-1 range as cellpose expects this

        masks_cytosol = model.eval(
            [input_image],
            rescale=self.rescale,
            normalize=self.normalize,
            diameter=self.diameter,
            flow_threshold=self.flow_threshold,
            cellprob_threshold=self.cellprob_threshold,
            channels=[2, 1],
        )[0]
        masks_cytosol = np.array(masks_cytosol)  # convert to array

        # manually delete model and perform gc to free up memory on GPU
        self._clear_cache(vars_to_delete=[model])

        # ensure edge classes are removed
        masks_cytosol = remove_edge_labels(masks_cytosol)

        #####
        ### Perform Filtering to remove too small/too large masks if applicable
        #####

        self._setup_filtering()

        if self.filter_size:
            masks_cytosol = self._perform_size_filtering(
                mask=masks_cytosol,
                thresholds=self.nucleus_thresholds,  # type: ignore
                confidence_interval=self.nucleus_confidence_interval,  # type: ignore
                mask_name="cytosol",
                log=True,
                debug=self.debug,
                input_image=input_image if self.debug else None,
            )

        masks_cytosol = masks_cytosol.reshape(masks_cytosol.shape[1:])  # add reshape to match shape to HDF5 shape

        return masks_cytosol

    def _execute_segmentation(self, input_image) -> None:
        total_time_start = timeit.default_timer()

        # transform input image
        input_image = self._transform_input_image(input_image)

        # check image dtype since cellpose expects int input images
        self._check_input_image_dtype(input_image)

        # execute segmentation
        start_segmentation = timeit.default_timer()
        cytosol_mask = self.cellpose_segmentation(input_image)
        stop_segmentation = timeit.default_timer()
        self.segmentation_time = stop_segmentation - start_segmentation

        # get final classes list
        all_classes = set(np.unique(cytosol_mask)) - {0}

        segmentation = self._finalize_segmentation_results(cytosol_mask)
        self._save_segmentation_sdata(segmentation, all_classes, masks=self.MASK_NAMES)

        # clean up memory
        self._clear_cache(vars_to_delete=[segmentation, all_classes])

        self.total_time = timeit.default_timer() - total_time_start

        return None


class ShardedCytosolOnlySegmentationCellpose(ShardedSegmentation):
    method = CytosolOnlySegmentationCellpose


class CytosolOnlySegmentationDownsamplingCellpose(CytosolOnlySegmentationCellpose):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _finalize_segmentation_results(self, cytosol_mask: np.ndarray) -> np.ndarray:
        cytosol_mask = self._rescale_downsampled_mask(cytosol_mask, "cytosol_segmentation")

        cytosol_mask = self._check_seg_dtype(mask=cytosol_mask, mask_name="cytosol")

        # combine masks into one stack
        segmentation = np.stack([cytosol_mask])

        return segmentation

    def _execute_segmentation(self, input_image) -> None:
        total_time_start = timeit.default_timer()

        # ensure the correct level is selected for the input image
        input_image = self._transform_input_image(input_image)

        # check image dtype since cellpose expects int input images
        self._check_input_image_dtype(input_image)

        # get the relevant channels for processing
        input_image = input_image[:2, :, :]
        self.input_image_size = input_image.shape

        # setup downsampling
        self._get_downsampling_parameters()
        self._calculate_padded_image_size(input_image)

        # downsample the image
        input_image = self._downsample_image(input_image)

        start_segmentation = timeit.default_timer()
        cytosol_mask = self.cellpose_segmentation(input_image)
        stop_segmentation = timeit.default_timer()
        self.segmentation_time = stop_segmentation - start_segmentation

        # currently no implemented filtering steps to remove nuclei outside of specific thresholds
        all_classes = set(np.unique(cytosol_mask)) - {0}

        segmentation = self._finalize_segmentation_results(cytosol_mask=cytosol_mask)  # type: ignore

        self._save_segmentation_sdata(segmentation, all_classes, masks=self.MASK_NAMES)
        self.total_time = timeit.default_timer() - total_time_start


class ShardedCytosolOnlySegmentationDownsamplingCellpose(ShardedSegmentation):
    method = CytosolOnlySegmentationDownsamplingCellpose



from scportrait.processing.images._image_processing import downsample_img_padding, percentile_normalization
from scportrait.processing.images._zstack_compression import EDF, maximum_intensity_projection



import os
import shutil

import pytest
import spatialdata as sd
import yaml
from spatialdata.datasets import blobs

from scportrait.data._datasets import dataset_1_omezarr
from scportrait.pipeline.project import Project


@pytest.fixture()
def sdata_path(tmp_path):
    sdata = blobs()

    # Write to temporary location
    sdata_path = tmp_path / "sdata.zarr"
    sdata.write(sdata_path)
    yield sdata_path
    shutil.rmtree(sdata_path)


@pytest.fixture()
def config_path(tmp_path):
    config_path = tmp_path / "config.yml"
    config = {"name": "Test segmentation"}

    with open(config_path, "w") as f:
        f.write(yaml.safe_dump(config))

    yield str(config_path)

    os.remove(config_path)


@pytest.mark.parametrize("image_name, segmentation_name", [("blobs_image", "blobs_labels")])
def test_project_load_input_from_sdata(sdata_path, config_path, tmp_path, image_name: str, segmentation_name: str):
    project_path = str(tmp_path / "scportrait/project/")

    project = Project(
        project_location=project_path,
        config_path=config_path,
        overwrite=True,
        debug=True,
    )

    project.load_input_from_sdata(sdata_path, input_image_name=image_name, cytosol_segmentation_name=segmentation_name)


@pytest.mark.parametrize("image_name, segmentation_name", [("blobs_multiscale_image", "blobs_multiscale_labels")])
def test_project_load_input_from_sdata_multiscale_image(
    sdata_path, config_path, tmp_path, image_name: str, segmentation_name: str
):
    project_path = str(tmp_path / "scportrait/project/")

    project = Project(
        project_location=project_path,
        config_path=config_path,
        overwrite=True,
        debug=True,
    )

    project.load_input_from_sdata(sdata_path, input_image_name=image_name, cytosol_segmentation_name=segmentation_name)


def test_project_load_from_omezarr(config_path, tmp_path):
    project_path = str(tmp_path / "scportrait/project/")
    omezarr_path = dataset_1_omezarr()

    project = Project(
        project_location=project_path,
        config_path=config_path,
        overwrite=True,
        debug=True,
    )

    project.load_input_from_omezarr(omezarr_path)
